# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
import numpy as np
import sparse
from sparse import COO as sp
from sparse import elemwise
import pandas as pd
import copy

sp.isnan = np.isnan
sp.newaxis = np.newaxis
sp.nan = np.array([1.0, np.nan])[-1]
sp.testing = np.testing
sp.nansum = sparse.nansum
sp.nanmin = sparse.nanmin
sp.nanmax = sparse.nanmax
sp.concatenate = sparse.concatenate
sp.diagonal = sparse.diagonal
sp.zeros = sparse.zeros
sp.testing.assert_array_equal = np.testing.assert_equal
sp.sqrt = np.sqrt
sp.log = np.log
sp.exp = np.exp
sp.abs = np.abs


def nan_to_num(a):
    if type(a) in [int, float, np.int64, np.float64]:
        return np.nan_to_num(a)
    if hasattr(a, "fill_value"):
        a = a.copy()
        a.data[np.isnan(a.data)] = 0.0
    return sp(coords=a.coords, data=a.data, fill_value=0.0, shape=a.shape)


def ones(*args, **kwargs):
    return sp(np.ones(*args, **kwargs), fill_value=sp.nan)


def nansum(a, axis=None, keepdims=None, *args, **kwargs):
    return sp(data=a.data, coords=a.coords, fill_value=0.0, shape=a.shape).sum(
        axis=axis, keepdims=keepdims, *args, **kwargs
    )
sp.nansum = nansum


def nanmean(a, axis=None, keepdims=None, *args, **kwargs):
    n = sp.nansum(a, axis=axis, keepdims=keepdims)
    d = sp.nansum(sp.nan_to_num(a) != 0, axis=axis, keepdims=keepdims).astype(n.dtype)
    n = sp(data=n.data, coords=n.coords, fill_value=np.nan, shape=n.shape)
    d = sp(data=d.data, coords=d.coords, fill_value=np.nan, shape=d.shape)
    out = n / d
    return sp(data=out.data, coords=out.coords, fill_value=0, shape=out.shape)

def array(a, *args, **kwargs):
    if kwargs.get("fill_value", None) is not None:
        fill_value = kwargs.pop("fill_value")
    else:
        fill_value = sp.nan
    if type(a) == sp:
        return sp(a, *args, **kwargs, fill_value=fill_value)
    else:
        return sp(np.array(a, *args, **kwargs), fill_value=fill_value)


def arange(*args, **kwargs):
    return sparse.COO.from_numpy(np.arange(*args, **kwargs))


def where(*args, **kwargs):
    return elemwise(np.where, *args, **kwargs)


def cumprod(a, axis=None, dtype=None, out=None):
    return array(np.cumprod(a.todense(), axis=axis, dtype=dtype, out=out))


def floor(x, *args, **kwargs):
    x.data = np.floor(x.data)
    return x



sp.minimum = np.minimum
sp.maximum = np.maximum
sp.floor = floor
sp.where = where
sp.arange = arange
sp.array = array
sp.nan_to_num = nan_to_num
sp.ones = ones
sp.cumprod = cumprod
sp.nanmean = nanmean



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
import pandas as pd
import numpy as np
from chainladder.utils.cupy import cp
from chainladder.utils.sparse import sp
import dill
import json
import os
import copy
from patsy import dmatrix
from sklearn.base import BaseEstimator, TransformerMixin
from typing import Iterable, Union
from io import StringIO


def load_sample(key: str, *args, **kwargs):
    """Function to load datasets included in the chainladder package.

    Parameters
    ----------
    key: str
        The name of the dataset, e.g. RAA, ABC, UKMotor, GenIns, etc.

    Returns
    -------
        pandas.DataFrame of the loaded dataset.

    """
    from chainladder import Triangle

    path = os.path.dirname(os.path.abspath(__file__))
    origin = "origin"
    development = "development"
    columns = ["values"]
    index = None
    cumulative = True
    if key.lower() in ["mcl", "usaa", "quarterly", "auto", "usauto", "tail_sample"]:
        columns = ["incurred", "paid"]
    if key.lower() == "clrd":
        origin = "AccidentYear"
        development = "DevelopmentYear"
        index = ["GRNAME", "LOB"]
        columns = [
            "IncurLoss",
            "CumPaidLoss",
            "BulkLoss",
            "EarnedPremDIR",
            "EarnedPremCeded",
            "EarnedPremNet",
        ]
    if key.lower() == "berqsherm":
        origin = "AccidentYear"
        development = "DevelopmentYear"
        index = ["LOB"]
        columns = ["Incurred", "Paid", "Reported", "Closed"]
    if key.lower() == "xyz":
        origin = "AccidentYear"
        development = "DevelopmentYear"
        columns = ["Incurred", "Paid", "Reported", "Closed", "Premium"]
    if key.lower() in ["liab", "auto"]:
        index = ["lob"]
    if key.lower() in ["cc_sample", "ia_sample"]:
        columns = ["loss", "exposure"]
    if key.lower() in ["prism"]:
        columns = ["reportedCount", "closedPaidCount", "Paid", "Incurred"]
        index = ["ClaimNo", "Line", "Type", "ClaimLiability", "Limit", "Deductible"]
        origin = "AccidentDate"
        development = "PaymentDate"
        cumulative = False
    df = pd.read_csv(os.path.join(path, "data", key.lower() + ".csv"))

    return Triangle(
        df,
        origin=origin,
        development=development,
        index=index,
        columns=columns,
        cumulative=cumulative,
        *args,
        **kwargs
    )


def read_pickle(path):
    with open(path, "rb") as pkl:
        return dill.load(pkl)


def read_json(json_str, array_backend=None):
    from chainladder import Triangle
    from chainladder.workflow import Pipeline

    if array_backend is None:
        from chainladder import options

        array_backend = options.ARRAY_BACKEND
    json_dict = json.loads(json_str)
    if type(json_dict) is list:
        import chainladder as cl

        return Pipeline(
            steps=[
                (
                    item["name"],
                    cl.__dict__[item["__class__"]]().set_params(**item["params"]),
                )
                for item in json_dict
            ]
        )
    elif "metadata" in json_dict.keys():
        j = json.loads(json_str)
        y = pd.read_json(StringIO(j["data"]), orient="split", date_unit="ns")
        y["origin"] = pd.to_datetime(y["origin"])
        y.columns = [c if c != "valuation" else "development" for c in y.columns]
        y["development"] = pd.to_datetime(y["development"])
        index = list(y.columns[: list(y.columns).index("origin")])
        columns = list(y.columns[list(y.columns).index("development") + 1 :])
        tri = Triangle(
            y,
            origin="origin",
            development="development",
            index=index,
            columns=columns,
            pattern=json.loads(j["metadata"])["is_pattern"],
            cumulative=False,
        )
        if json.loads(j["metadata"])["is_val_tri"]:
            tri = tri.dev_to_val()
        if json.loads(j["metadata"])["is_cumulative"]:
            tri = tri.incr_to_cum()
        tri = tri[json.loads(j["metadata"])["columns"]].sort_index()
        if "sub_tris" in json_dict.keys():
            for k, v in json_dict["sub_tris"].items():
                setattr(tri, k, read_json(v, array_backend))
                setattr(getattr(tri, k), "origin_grain", tri.origin_grain)
                setattr(getattr(tri, k), "development_grain", tri.development_grain)
        if "dfs" in json_dict.keys():
            for k, v in json_dict["dfs"].items():
                df = pd.read_json(v)
                if len(df.columns) == 1:
                    df = df.iloc[:, 0]
                setattr(tri, k, df)
        if array_backend:
            return tri.set_backend(array_backend)
        else:
            return tri
    else:
        import chainladder as cl

        return cl.__dict__[json_dict["__class__"]]().set_params(**json_dict["params"])


def parallelogram_olf(
    values,
    date,
    start_date=None,
    end_date=None,
    grain="Y",
    approximation_grain="M",
    vertical_line=False,
):
    """Parallelogram approach to on-leveling."""
    date = pd.to_datetime(date)
    if not start_date:
        start_date = "{}-01-01".format(date.min().year)
    if not end_date:
        end_date = "{}-12-31".format(date.max().year)
    start_date = pd.to_datetime(start_date) - pd.tseries.offsets.DateOffset(days=1)

    date_freq = {
        "M": "MS",
        "D": "D",
    }
    if approximation_grain not in ['M', 'D']:
        raise ValueError("approximation_grain must be " "M" " or " "D" "")
    date_idx = pd.date_range(
        start_date - pd.tseries.offsets.DateOffset(years=1),
        end_date,
        freq=date_freq[approximation_grain],
    )

    rate_changes = pd.Series(np.array(values), np.array(date))
    rate_changes = rate_changes.reindex(date_idx, fill_value=0)

    cum_rate_changes = np.cumprod(1 + rate_changes.values)
    cum_rate_changes = pd.Series(cum_rate_changes, rate_changes.index)
    crl = cum_rate_changes.iloc[-1]

    cum_avg_rate_non_leaps = cum_rate_changes
    cum_avg_rate_leaps = cum_rate_changes

    if not vertical_line:
        rolling_num = {
            "M": 12,
            "D": 365,
        }

        cum_avg_rate_non_leaps = cum_rate_changes.rolling(
            rolling_num[approximation_grain]
        ).mean()
        cum_avg_rate_non_leaps = (
            cum_avg_rate_non_leaps + cum_avg_rate_non_leaps.shift(1).values
        ) / 2

        cum_avg_rate_leaps = cum_rate_changes.rolling(
            rolling_num[approximation_grain] + 1
        ).mean()
        cum_avg_rate_leaps = (
            cum_avg_rate_leaps + cum_avg_rate_leaps.shift(1).values
        ) / 2

    dropdates_num = {
        "M": 12,
        "D": 366,
    }
    cum_avg_rate_non_leaps = cum_avg_rate_non_leaps.iloc[
        dropdates_num[approximation_grain] :
    ]
    cum_avg_rate_leaps = cum_avg_rate_leaps.iloc[
        dropdates_num[approximation_grain] + 1 :
    ]

    fcrl_non_leaps = (
        cum_avg_rate_non_leaps.groupby(cum_avg_rate_non_leaps.index.to_period(grain))
        .mean()
        .reset_index()
    )
    fcrl_non_leaps.columns = ["Origin", "OLF"]
    fcrl_non_leaps["Origin"] = fcrl_non_leaps["Origin"].astype(str)
    fcrl_non_leaps["OLF"] = crl / fcrl_non_leaps["OLF"]

    fcrl_leaps = (
        cum_avg_rate_leaps.groupby(cum_avg_rate_leaps.index.to_period(grain))
        .mean()
        .reset_index()
    )
    fcrl_leaps.columns = ["Origin", "OLF"]
    fcrl_leaps["Origin"] = fcrl_leaps["Origin"].astype(str)
    fcrl_leaps["OLF"] = crl / fcrl_leaps["OLF"]

    combined = fcrl_non_leaps.join(fcrl_leaps, lsuffix="_non_leaps", rsuffix="_leaps")
    combined["is_leap"] = pd.to_datetime(
        combined["Origin_non_leaps"], format="%Y" + ("-%M" if grain == "M" else "")
    ).dt.is_leap_year
    

    if approximation_grain == "M":
        combined["final_OLF"] = combined["OLF_non_leaps"]
    else:
        combined["final_OLF"] = np.where(
            combined["is_leap"], combined["OLF_leaps"], combined["OLF_non_leaps"]
        )

    combined.drop(
        ["OLF_non_leaps", "Origin_leaps", "OLF_leaps", "is_leap"],
        axis=1,
        inplace=True,
    )
    combined.columns = ["Origin", "OLF"]

    return combined.set_index("Origin")


def set_common_backend(objs):
    from chainladder import options

    priority = options.ARRAY_PRIORITY
    backend = priority[np.min([priority.index(i.array_backend) for i in objs])]
    return [i.set_backend(backend) for i in objs]


def concat(
    objs: Iterable,
    axis: Union[int, str],
    ignore_index: bool = False,
    sort: bool = False,
):
    """Concatenate Triangle objects along a particular axis.

    Parameters
    ----------
    objs: list or tuple
        A list or tuple of Triangle objects to concat. All non-concat axes must
        be identical and all elements of the concat axes must be unique.
    axis: string or int
        The axis to concatenate along.
    ignore_index: bool, default False
        If True, do not use the index values along the concatenation axis. The
        resulting axis will be labeled 0, …, n - 1. This is useful if you are
        concatenating objects where the concatenation axis does not have
        meaningful indexing information. Note the index values on the other
        axes are still respected in the join.

    Returns
    -------
    Updated triangle
    """
    if type(objs) not in (list, tuple):
        raise TypeError("objects to be concatenated must be in a list or tuple")
    if type(objs) is tuple:
        objs = list(objs)
    if len(objs) == 0:
        raise ValueError("objs must contain at least one element")
    xp = objs[0].get_array_module()
    axis = objs[0]._get_axis(axis)

    if axis != 1:
        all_columns = []
        for obj in objs:
            for col in obj.columns:
                if col not in all_columns:
                    all_columns.append(col)
        for num, obj in enumerate(objs):
            for col in all_columns:
                if col not in obj.columns:
                    obj = copy.deepcopy(obj)
                    obj[col] = xp.nan
                    objs[num] = obj
    objs = set_common_backend(objs)
    mapper = {0: "kdims", 1: "vdims", 2: "odims", 3: "ddims"}
    for k in mapper.keys():
        if k != axis and k != 1:  # All non-concat axes must be identical
            a = np.array([getattr(obj, mapper[k]) for obj in objs])
            assert np.all(a == a[0])
        else:  # All elements of concat axis must be unique
            if ignore_index:
                new_axis = np.arange(
                    np.sum([len(getattr(obj, mapper[axis])) for obj in objs])
                )
                new_axis = new_axis[:, None] if axis == 0 else new_axis
            else:
                new_axis = np.concatenate([getattr(obj, mapper[axis]) for obj in objs])
            if axis == 0:
                assert len(pd.DataFrame(new_axis).drop_duplicates()) == len(new_axis)
            else:
                assert len(new_axis) == len(set(new_axis))
    out = copy.deepcopy(objs[0])
    out.values = xp.concatenate([obj.values for obj in objs], axis=axis)
    setattr(out, mapper[axis], new_axis)
    if ignore_index and axis == 0:
        out.key_labels = ["Index"]
    out.valuation_date = pd.Series([obj.valuation_date for obj in objs]).max()
    if out.ddims.dtype == "datetime64[ns]" and type(out.ddims) == np.ndarray:
        out.ddims = pd.DatetimeIndex(out.ddims)
    out._set_slicers()
    if sort:
        return out.sort_axis(axis)
    else:
        return out


def num_to_value(arr, value):
    """Function that turns all zeros to nan values in an array"""
    backend = arr.__class__.__module__.split(".")[0]
    if backend == "sparse":
        if arr.fill_value == 0 or sp.isnan(arr.fill_value):
            arr.coords = arr.coords[:, arr.data != 0]
            arr.data = arr.data[arr.data != 0]

            arr = sp(
                coords=arr.coords, data=arr.data, fill_value=sp.nan, shape=arr.shape
            )
        else:
            arr = sp(num_to_nan(np.nan_to_num(arr.todense())), fill_value=value)
    else:
        arr[arr == 0] = value
    return arr


def num_to_nan(arr):
    """Function that turns all zeros to nan values in an array"""
    from chainladder import Triangle

    xp = Triangle.get_array_module(None, arr=arr)

    return num_to_value(arr, xp.nan)


def minimum(x1, x2):
    return x1.minimum(x2)


def maximum(x1, x2):
    return x1.maximum(x2)


class PatsyFormula(BaseEstimator, TransformerMixin):
    """A sklearn-style Transformer for patsy formulas.

    PatsyFormula allows for R-style formula preprocessing of the ``design_matrix``
    of a machine learning algorithm. It's particularly useful with the `DevelopmentML`
    and `TweedieGLM` estimators.

    Parameters
    ----------

    formula: str
        A string representation of the regression model X features.

    Attributes
    ----------
    design_info_:
        The patsy instructions for generating the design_matrix, X.

    """

    def __init__(self, formula=None):
        self.formula = formula

    def _check_X(self, X):
        from chainladder.core import Triangle

        if isinstance(X, Triangle):
            raise AttributeError("X must be a pandas dataframe, not a Triangle")

    def fit(self, X, y=None, sample_weight=None):
        self._check_X(X)
        self.design_info_ = dmatrix(self.formula, X).design_info
        return self

    def transform(self, X):
        self._check_X(X)
        return dmatrix(self.design_info_, X)


def model_diagnostics(model, name=None, groupby=None):
    """A helper function that summarizes various vectors of an
    IBNR model as columns of a Triangle

    Parameters
    ----------
    model:
        A chainladder IBNR estimator or Pipeline
    name:
        An alias to give the model. This will be added to the index of the return
        triangle.
    groupby:
        The index level at which the model should be summarized

    Returns
    -------
    Triangle up select origin vectors, IBNR, ultimate, Latest diagonal, etc.
    """
    from chainladder import Pipeline

    if isinstance(model, Pipeline):
        obj = copy.deepcopy(model.steps[-1][-1])
    else:
        obj = copy.deepcopy(model)
    if groupby is not None:
        obj.X_ = obj.X_.groupby(groupby).sum().cum_to_incr()
        obj.ultimate_ = obj.ultimate_.groupby(groupby).sum()
        if hasattr(obj, "expectation_"):
            obj.expectation_ = obj.expectation_.groupby(groupby).sum()
    else:
        obj.X_ = obj.X_.incr_to_cum()
    val = obj.X_.valuation
    latest = obj.X_.sum("development")
    run_off = obj.full_expectation_.iloc[..., :-1].dev_to_val().cum_to_incr()
    run_off = run_off[run_off.development > str(obj.X_.valuation_date)]
    run_off = run_off.iloc[
        ..., : {"M": 12, "S": 6, "Q": 4, "Y": 1}[obj.X_.development_grain]
    ]

    triangles = []
    for col in obj.ultimate_.columns:
        idx = latest.index
        idx["Measure"] = col
        idx["Model"] = obj.__class__.__name__ if name is None else name
        idx = idx[list(idx.columns[-2:]) + list(idx.columns[:-2])]
        out = latest[col].rename("columns", ["Latest"])
        if obj.X_.development_grain in ["M"]:
            out["Month Incremental"] = obj.X_[col][val == obj.X_.valuation_date].sum(
                "development"
            )
        if (
            obj.X_.development_grain in ["M", "Q"]
            and pd.Period(out.valuation_date, freq="Q").to_timestamp(how="s")
            > val.min()
        ):
            out["Quarter Incremental"] = (
                obj.X_
                - obj.X_[
                    val
                    < pd.Period(out.valuation_date, freq="Q")
                    .to_timestamp(how="s")
                    .strftime("%Y-%m")
                ]
            ).sum("development")[col]
        else:
            out["Quarter Incremental"] = 0
        if pd.Period(out.valuation_date, freq="Y").to_timestamp(how="s") > val.min():
            out["Year Incremental"] = (
                obj.X_ - obj.X_[val < str(obj.X_.valuation_date.year)]
            ).sum("development")[col]
        else:
            out["Year Incremental"] = 0
        out["IBNR"] = obj.ibnr_[col]
        out["Ultimate"] = obj.ultimate_[col]
        for i in range(run_off.shape[-1]):
            out["Run Off " + str(i + 1)] = run_off[col].iloc[..., i]
        if hasattr(obj, "expectation_"):
            out["Apriori"] = (
                obj.expectation_
                if obj.expectation_.shape[1] == 1
                else obj.expectation_[col]
            )
        out.index = idx
        triangles.append(out)
    return concat(triangles, 0)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
import pandas as pd
import numpy as np
from chainladder.utils.utility_functions import num_to_nan, concat
from chainladder.core.pandas import TriangleGroupBy
from chainladder.utils.sparse import sp

try:
    import dask.bag as db
except:
    db = None

class TriangleDunders:
    """ Class that implements the dunder (double underscore) methods for the
        Triangle class
    """

    def _validate_arithmetic(self, other):
        """ Common functionality BEFORE arithmetic operations """
        if isinstance(other, TriangleDunders):
            obj, other = self._compatibility_check(self, other)
            if obj.is_pattern != other.is_pattern:
                obj.is_pattern = other.is_pattern = False
            obj.valuation_date = max(obj.valuation_date, other.valuation_date)
            obj, other = self._prep_columns(obj, other)
            obj, other = self._prep_origin_development(obj, other)
            obj, other = self._prep_index(obj, other)
            if isinstance(other, TriangleDunders):
                other = other.values
        else:
            if isinstance(other, np.ndarray) and self.array_backend != 'numpy':
                obj = self.copy()
                other = obj.get_array_module().array(other)
            elif isinstance(other, sp) and self.array_backend != 'sparse':
                obj = self.set_backend('sparse')
            else:
                obj = self.copy()
        return obj, other

    def _arithmetic_cleanup(self, obj):
        """ Common functionality AFTER arithmetic operations """
        obj.values = obj.values * obj.get_array_module().nan_to_num(obj.nan_triangle)
        obj.values = num_to_nan(obj.values)
        return obj

    def _compatibility_check(self, x, y):
        from chainladder.utils.utility_functions import set_common_backend

        x, y = set_common_backend([x, y])
        if (
            x.origin_grain != y.origin_grain
            or (x.development_grain != y.development_grain and 
                min(x.shape[-1], y.shape[-1]) > 1)
        ):
            raise ValueError(
                "Triangle arithmetic requires both triangles to be the same grain."
            )
        return x, y

    def _prep_index(self, x, y):
        """ Preps index and column axes for arithmetic """
        if x.kdims.shape[0] == 1 and y.kdims.shape[0] > 1:
            # Broadcast x to y
            x.kdims = y.kdims
            x.key_labels = y.key_labels
            return x, y
        if x.kdims.shape[0] > 1 and y.kdims.shape[0] == 1:
            # Broadcast y to x
            y.kdims = x.kdims
            y.key_labels = x.key_labels
            return x, y
        if x.kdims.shape[0] == y.kdims.shape[0] == 1 and x.key_labels != y.key_labels:
            # Broadcast to the triangle with a larger multi-index
            kdims = x.kdims if len(x.key_labels) > len(y.key_labels) else y.kdims
            y.kdims = x.kdims = kdims
            return x, y
        a, b = set(x.key_labels), set(y.key_labels)
        common = a.intersection(b)
        if common in [a, b] and (a != b or (a == b and x.kdims.shape[0] != y.kdims.shape[0])):
            # If index labels are subset of other triangle index labels
            x = x.groupby(list(common))
            y = y.groupby(list(common))
            return x, y
        if common not in [a, b]:
            raise ValueError('Index broadcasting is ambiguous between', str(a), 'and', str(b))
        if (
            x.key_labels == y.key_labels
            and x.kdims.shape[0] == y.kdims.shape[0]
            and y.kdims.shape[0] > 1
            and not x.kdims is y.kdims
            and not x.index.equals(y.index)
        ):
            # Make sure exact but unsorted index labels works
            x = x.sort_index()
            try:
                y = y.loc[x.index]
            except:
                x = x.groupby(list(common))
                y = y.groupby(list(common))
        return x, y

    def _prep_columns(self, x, y):
        x_backend, y_backend = x.array_backend, y.array_backend
        if len(x.columns) == 1 and len(y.columns) > 1:
            x.vdims = y.vdims
        elif len(y.columns) == 1 and len(x.columns) > 1:
            y.vdims = x.vdims
        elif len(y.columns) == 1 and len(x.columns) == 1 and x.columns != y.columns:
            y.vdims = x.vdims
        elif x.shape[1] == y.shape[1] and np.all(x.columns == y.columns):
            pass
        else:
            col_union = list(x.columns) + [
                item for item in y.columns if item not in x.columns
            ]
            for item in [item for item in col_union if item not in x.columns]:
                x[item] = 0
            x = x[col_union]
            for item in [item for item in col_union if item not in y.columns]:
                y[item] = 0
            y = y[col_union]
        x, y = (
            x.set_backend(x_backend, inplace=True),
            y.set_backend(y_backend, inplace=True),
        )
        return x, y

    def _prep_origin_development(self, obj, other):
        xp = obj.get_array_module()
        is_broadcastable = all(
            (m == n) or (m == 1) or (n == 1)
            for m, n in zip(obj.shape[-2:][::-1], other.shape[-2:][::-1]))
        if (len(obj.odims), len(obj.ddims)) == (len(other.odims), len(other.ddims)):
            if np.all(obj.ddims != other.ddims) and len(obj.ddims) > 1:
                is_broadcastable = False
            if np.all(obj.odims != other.odims) and len(obj.odims) > 1:
                is_broadcastable = False
        if len(other.odims) == 1 and len(obj.odims) > 1:
            other.odims = obj.odims
            other.origin_grain = obj.origin_grain
        elif len(obj.odims) == 1 and len(other.odims) > 1:
            obj.odims = other.odims
            obj.origin_grain = other.origin_grain
        if len(other.ddims) == 1 and len(obj.ddims) > 1:
            other.ddims = obj.ddims
            other.development_grain = obj.development_grain
        elif len(obj.ddims) == 1 and len(other.ddims) > 1:
            obj.ddims = other.ddims
            obj.development_grain = other.development_grain
        if not is_broadcastable:
            # If broadcasting doesn't work, union axes similar to pandas
            ddims = pd.concat(
                (
                    pd.Series(obj.ddims, index=obj.ddims),
                    pd.Series(other.ddims, index=other.ddims),
                ),
                axis=1,
            ).sort_index()
            odims = pd.concat(
                (
                    pd.Series(obj.odims, index=obj.odims),
                    pd.Series(other.odims, index=other.odims),
                ),
                axis=1,
            ).sort_index()
            o_arr0, o_arr1 = odims[0].isna().values, odims[1].isna().values
            d_arr0, d_arr1 = ddims[0].isna().values, ddims[1].isna().values
            # rol = right hand side, origin, lower
            rol = int(np.where(~o_arr1 == 1)[0].min())
            roh = int(np.where(~o_arr1 == 1)[0].max() + 1)
            rdl = int(np.where(~d_arr1 == 1)[0].min())
            rdh = int(np.where(~d_arr1 == 1)[0].max() + 1)
            lol = int(np.where(~o_arr0 == 1)[0].min())
            loh = int(np.where(~o_arr0 == 1)[0].max() + 1)
            ldl = int(np.where(~d_arr0 == 1)[0].min())
            ldh = int(np.where(~d_arr0 == 1)[0].max() + 1)
            if obj.array_backend != "sparse":
                other_arr = xp.zeros((other.shape[0], other.shape[1], len(odims), len(ddims)))
                other_arr[:] = xp.nan
                other_arr[:, :, rol:roh, rdl:rdh] = other.values
                obj_arr = xp.zeros((self.shape[0], self.shape[1], len(odims), len(ddims)))
                obj_arr[:] = xp.nan
                obj_arr[:, :, lol:loh, ldl:ldh] = obj.values
            else:
                obj_arr, other_arr = obj.values.copy(), other.values.copy()
                other_arr.coords[2] = other_arr.coords[2] + rol
                other_arr.coords[3] = other_arr.coords[3] + rdl
                obj_arr.coords[2] = obj_arr.coords[2] + lol
                obj_arr.coords[3] = obj_arr.coords[3] + ldl
                other_arr.shape = (other.shape[0], other.shape[1], len(odims), len(ddims))
                obj_arr.shape = (self.shape[0], self.shape[1], len(odims), len(ddims))
            obj.odims = np.array(odims.index)
            if type(obj.ddims) == pd.DatetimeIndex:
                obj.ddims = pd.DatetimeIndex(ddims.index)
            else:
                obj.ddims = np.array(ddims.index)
            obj.values = obj_arr
            other.values = other_arr
        return obj, other

    @staticmethod
    def _slice_or_nan(obj, other, k):
        """ Return a broadcastable slice or NaNs for missing slices """
        if k in obj.groups.indices.keys():
            return obj.obj.iloc[obj.groups.indices[k]]
        else:
            other = other.obj.iloc[other.groups.indices[k]]
            new_obj = obj.obj.iloc[:1] * 0
            labels = list(set(other.key_labels).intersection(set(new_obj.key_labels)))
            new_idx = other.index.set_index(labels).join(
                new_obj.index.set_index(labels)).reset_index()
            new_idx = new_idx[new_obj.key_labels].iloc[-1:]
            new_obj.kdims = new_idx.values
            new_obj.key_labels = list(new_idx.columns)
            return new_obj

    @staticmethod
    def _get_key_union(obj, other):
        return set(list(obj.groups.indices.keys()) +
                   list(other.groups.indices.keys()))

    def _arithmetic_mapper(self, obj, other, f):
        """ Use Dask if available, otherwise basic list comprehension """
        if db and obj.obj.array_backend == 'sparse':
            bag = db.from_sequence(self._get_key_union(obj, other))
            bag = bag.map(f, self, obj, other)
            c = bag.compute(scheduler='threads')
        else:
            c = [f(k, self, obj, other) for k in self._get_key_union(obj, other)]
        return concat(c, 0).sort_index()

    def __add__(self, other):
        obj, other = self._validate_arithmetic(other)
        if isinstance(obj, TriangleGroupBy):
            def f(k, self, obj, other):
                return (self._slice_or_nan(obj, other, k) +
                        self._slice_or_nan(other, obj, k))
            obj = self._arithmetic_mapper(obj, other, f)
        else:
            xp = obj.get_array_module()
            obj.values = xp.nan_to_num(obj.values) + xp.nan_to_num(other)
        return self._arithmetic_cleanup(obj)

    def __radd__(self, other):
        return self if other == 0 else self.__add__(other)

    def __sub__(self, other):
        obj, other = self._validate_arithmetic(other)
        if isinstance(obj, TriangleGroupBy):
            def f(k, self, obj, other):
                return (self._slice_or_nan(obj, other, k) -
                        self._slice_or_nan(other, obj, k))
            obj = self._arithmetic_mapper(obj, other, f)
        else:
            xp = obj.get_array_module()
            obj.values = xp.nan_to_num(obj.values) - xp.nan_to_num(other)
        return self._arithmetic_cleanup(obj)

    def __rsub__(self, other):
        obj, other = self._validate_arithmetic(other)
        xp = obj.get_array_module()
        obj.values = xp.nan_to_num(other) - xp.nan_to_num(obj.values)
        return self._arithmetic_cleanup(obj)

    def __len__(self):
        return self.shape[0]

    def __neg__(self):
        obj = self.copy()
        obj.values = -obj.values
        return obj

    def __pos__(self):
        return self

    def __abs__(self):
        obj = self.copy()
        obj.values = abs(obj.values)
        return obj

    def __mul__(self, other):
        obj, other = self._validate_arithmetic(other)
        if isinstance(obj, TriangleGroupBy):
            def f(k, self, obj, other):
                return (self._slice_or_nan(obj, other, k) *
                        self._slice_or_nan(other, obj, k))
            obj = self._arithmetic_mapper(obj, other, f)
        else:
            xp = obj.get_array_module()
            obj.values = obj.values * other
        return obj

    def __rmul__(self, other):
        return self if other == 1 else self.__mul__(other)

    def __pow__(self, other):
        obj, other = self._validate_arithmetic(other)
        if isinstance(obj, TriangleGroupBy):
            def f(k, self, obj, other):
                return (self._slice_or_nan(obj, other, k) **
                        self._slice_or_nan(other, obj, k))
            obj = self._arithmetic_mapper(obj, other, f)
        else:
            xp = obj.get_array_module()
            obj.values = xp.nan_to_num(obj.values) ** other
        return obj

    def __round__(self, other):
        obj = self.copy()
        xp = obj.get_array_module()
        obj.values = xp.round(obj.values, other)
        return obj

    def __truediv__(self, other):
        obj, other = self._validate_arithmetic(other)
        if isinstance(obj, TriangleGroupBy):
            def f(k, self, obj, other):
                return (self._slice_or_nan(obj, other, k) /
                        self._slice_or_nan(other, obj, k))
            obj = self._arithmetic_mapper(obj, other, f)
        else:
            xp = obj.get_array_module()
            obj.values = obj.values / other
        return obj

    def __rtruediv__(self, other):
        obj = self.copy()
        obj.values = other / self.values
        obj.values = num_to_nan(obj.values)
        return obj

    def __eq__(self, other):
        if not isinstance(other, TriangleDunders):
            return False
        from chainladder import options

        backend = options.ARRAY_PRIORITY[
            min(
                [
                    options.ARRAY_PRIORITY.index(x)
                    for x in [self.array_backend, other.array_backend]
                ]
            )
        ]
        x, y = self.set_backend(backend), other.set_backend(backend)
        xp = x.get_array_module()
        return xp.all(xp.nan_to_num(x.values) == xp.nan_to_num(y.values))

    def __contains__(self, value):
        return self.__dict__.get(value, None) is not None

    def __lt__(self, value):
        obj = self.copy()
        xp = self.get_array_module()
        obj.values = xp.nan_to_num(obj.values) < xp.nan_to_num(value)
        return obj

    def __le__(self, value):
        obj = self.copy()
        xp = self.get_array_module()
        obj.values = xp.nan_to_num(obj.values) < xp.nan_to_num(value)
        return obj



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
import numpy as np
import pandas as pd
import warnings
from sklearn.base import BaseEstimator, TransformerMixin
from chainladder.utils import WeightedRegression
from chainladder.utils.utility_functions import num_to_nan
from chainladder.core.io import EstimatorIO
from chainladder.core.common import Common


class DevelopmentBase(BaseEstimator, TransformerMixin, EstimatorIO, Common):
    def fit(self, X, y=None, sample_weight=None):
        average_ = self._validate_assumption(y, self.average, axis=3)
        self.average_ = average_.flatten()
        exponent = self.xp.array(
            [{"regression": 0, "volume": 1, "simple": 2}[x] for x in average_[0, 0, 0]]
        )
        exponent = self.xp.nan_to_num(exponent * (y * 0 + 1))
        w = num_to_nan(sample_weight / (X ** (exponent)))
        self.params_ = WeightedRegression(axis=2, thru_orig=True, xp=self.xp).fit(
            X, y, w
        )
        return self

    def _set_fit_groups(self, X):
        """Used for assigning group_index in fit"""
        backend = "numpy" if X.array_backend in ["sparse", "numpy"] else "cupy"
        if self.groupby is None:
            return X.set_backend(backend)
        if callable(self.groupby) or type(self.groupby) in [list, str, pd.Series]:
            return X.groupby(self.groupby).sum().set_backend(backend)
        else:
            raise ValueError("Cannot determine groupings.")

    def _set_transform_groups(self, X):
        """Used for assigning group_index in transform"""
        if self.groupby is None:
            return X.index
        else:
            indices = X.groupby(self.groupby).groups.indices
        return (
            pd.Series(
                {vi: k for k, v in indices.items() for vi in v},
                name=self.ldf_.key_labels[0],
            )
            .sort_index()
            .to_frame()
        )

    def _assign_n_periods_weight(self, X, n_periods):
        """Used to apply the n_periods weight"""

        def _assign_n_periods_weight_int(X, n_periods):
            xp = X.get_array_module()
            val_offset = {
                "Y": {"Y": 1},
                "S": {"Y": 2, "S": 1},
                "Q": {"Y": 4, "S": 2, "Q": 1},
                "M": {"Y": 12, "S": 6, "Q": 3, "M": 1},
            }
            if n_periods < 1 or n_periods >= X.shape[-2] - 1:
                return X.values * 0 + 1
            else:
                val_date_min = X.valuation[X.valuation <= X.valuation_date]
                val_date_min = val_date_min.drop_duplicates().sort_values()
                z = -n_periods * val_offset[X.development_grain][X.origin_grain] - 1
                val_date_min = val_date_min[z]
                w = X[X.valuation >= val_date_min]
                return xp.nan_to_num((w / w).values) * X.nan_triangle

        xp = X.get_array_module()

        dict_map = {
            item: _assign_n_periods_weight_int(X, item)
            for item in set(n_periods.flatten())
        }

        conc = [
            dict_map[item][..., num : num + 1]
            for num, item in enumerate(n_periods.flatten())
        ]
        return xp.concatenate(tuple(conc), -1)

    def _drop_adjustment(self, X, link_ratio):
        weight = X.nan_triangle[:, :-1]

        if self.drop is not None:
            weight = weight * self._drop(X)

        if self.drop_valuation is not None:
            weight = weight * self._drop_valuation(X)

        if (self.drop_high is not None) | (self.drop_low is not None):
            n_periods_ = self._validate_assumption(X, self.n_periods, axis=3)[
                0, 0, 0, :-1
            ]

            w_ = self._assign_n_periods_weight(X, n_periods_)
            w_ = w_.astype("float")
            w_[w_ == 0] = np.nan

            link_ratio_w_ = link_ratio * w_

            weight = weight * self._drop_n(
                self.drop_high,
                self.drop_low,
                X,
                link_ratio_w_,
                self.preserve,
            )

        if (self.drop_above != np.inf) | (self.drop_below != 0.00):
            weight = weight * self._drop_x(
                self.drop_above, self.drop_below, X, link_ratio, self.preserve
            )

        return weight

    # for drop_high and drop_low
    def _drop_n(self, drop_high, drop_low, X, link_ratio, preserve):
        # this is safe because each triangle by index and column has
        link_ratios_len = link_ratio.shape[3]

        def drop_array_helper(drop_type):
            drop_type_array = np.array(link_ratios_len * [0])

            if drop_type is None:
                return np.array(link_ratios_len * [0])
            else:
                # only a single parameter is provided
                if isinstance(drop_type, int):
                    drop_type_array = np.array(link_ratios_len * [drop_type])
                elif isinstance(drop_type, bool):
                    drop_type_array = np.array(link_ratios_len * int(drop_type))

                # an array of parameters is provided
                else:
                    for index in range(len(drop_type)):
                        drop_type_array[index] = int(drop_type[index])

                # convert boolean to ints (1s)
                for index in range(len(drop_type_array)):
                    if isinstance(drop_type_array[index], bool):
                        drop_type_array[index] = int(drop_type_array[index] == True)
                    else:
                        drop_type_array[index] = drop_type_array[index]

                return drop_type_array

        # explicitly setting up 3D arrays for drop parameters to avoid broadcasting bugs
        drop_high_array = np.zeros(
            (link_ratio.shape[0], link_ratio.shape[1], link_ratios_len)
        )
        drop_high_array[:, :, :] = drop_array_helper(drop_high)
        drop_low_array = np.zeros(
            (link_ratio.shape[0], link_ratio.shape[1], link_ratios_len)
        )
        drop_low_array[:, :, :] = drop_array_helper(drop_low)
        n_period_array = np.zeros(
            (link_ratio.shape[0], link_ratio.shape[1], link_ratios_len)
        )
        n_period_array[:, :, :] = drop_array_helper(self.n_periods)
        preserve_array = np.zeros(
            (link_ratio.shape[0], link_ratio.shape[1], link_ratios_len)
        )
        preserve_array[:, :, :] = drop_array_helper(preserve)

        # operationalizing the -1 option for n_period
        n_period_array = np.where(n_period_array == -1, link_ratios_len, n_period_array)

        # ranking factors by itself and volume
        link_ratio_ranks = np.lexsort((X.values[..., :-1], link_ratio), axis=2).argsort(
            axis=2
        )

        # setting up default return
        weights = ~np.isnan(link_ratio.transpose((0, 1, 3, 2)))

        # counting valid factors
        ldf_count = weights.sum(axis=3)

        # applying n_period
        ldf_count_n_period = np.where(
            ldf_count > n_period_array, n_period_array, ldf_count
        )

        # applying drop_high and drop_low
        max_rank_unpreserve = ldf_count_n_period - drop_high_array
        min_rank_unpreserve = drop_low_array

        # applying preserve
        warning_flag = np.any(max_rank_unpreserve - min_rank_unpreserve < preserve)
        max_rank = np.where(
            max_rank_unpreserve - min_rank_unpreserve < preserve,
            ldf_count_n_period,
            max_rank_unpreserve,
        )
        min_rank = np.where(
            max_rank_unpreserve - min_rank_unpreserve < preserve, 0, min_rank_unpreserve
        )

        index_array_weights = (
            link_ratio_ranks.transpose((0, 1, 3, 2))
            < max_rank.reshape(
                max_rank.shape[0], max_rank.shape[1], max_rank.shape[2], 1
            )
        ) & (
            link_ratio_ranks.transpose((0, 1, 3, 2))
            > min_rank.reshape(
                min_rank.shape[0], min_rank.shape[1], min_rank.shape[2], 1
            )
            - 1
        )

        weights = index_array_weights

        if warning_flag:
            if preserve == 1:
                warning = (
                    "Some exclusions have been ignored. At least "
                    + str(preserve)
                    + " (use preserve = ...)"
                    + " link ratio(s) is required for development estimation."
                )
            else:
                warning = (
                    "Some exclusions have been ignored. At least "
                    + str(preserve)
                    + " link ratio(s) is required for development estimation."
                )
            warnings.warn(warning)

        return weights.transpose((0, 1, 3, 2))

    # for drop_above and drop_below
    def _drop_x(self, drop_above, drop_below, X, link_ratio, preserve):
        # this is safe because each triangle by index and column has
        link_ratios_len = link_ratio.shape[3]

        def drop_array_helper(drop_type, default_value):
            drop_type_array = np.array(link_ratios_len * [default_value])

            # only a single parameter is provided
            if isinstance(drop_type, int):
                drop_type_array = np.array(link_ratios_len * [float(drop_type)])
            elif isinstance(drop_type, float):
                drop_type_array = np.array(link_ratios_len * [drop_type])

            # an array of parameters is provided
            else:
                for index in range(len(drop_type)):
                    drop_type_array[index] = float(drop_type[index])

            return drop_type_array

        # explicitly setting up 3D arrays for drop parameters to avoid broadcasting bugs
        drop_above_array = np.zeros(
            (link_ratio.shape[0], link_ratio.shape[1], link_ratios_len)
        )
        drop_above_array[:, :, :] = drop_array_helper(drop_above, np.inf)
        drop_below_array = np.zeros(
            (link_ratio.shape[0], link_ratio.shape[1], link_ratios_len)
        )
        drop_below_array[:, :, :] = drop_array_helper(drop_below, 0.0)
        preserve_array = np.zeros(
            (link_ratio.shape[0], link_ratio.shape[1], link_ratios_len)
        )
        preserve_array[:, :, :] = drop_array_helper(preserve, preserve)

        # transposing
        link_ratio_T = link_ratio.transpose((0, 1, 3, 2))

        # setting up default return
        weights = ~np.isnan(link_ratio_T)

        # dropping
        index_array_weights = (link_ratio_T < drop_above_array[..., None]) & (
            link_ratio_T > drop_below_array[..., None]
        )

        # counting remaining factors
        ldf_count = index_array_weights.sum(axis=3)

        # applying preserve
        warning_flag = np.any(ldf_count < preserve_array)
        weights = np.where(
            ldf_count[..., None] < preserve_array[..., None],
            weights,
            index_array_weights,
        )

        if warning_flag:
            if preserve == 1:
                warning = (
                    "Some exclusions have been ignored. At least "
                    + str(preserve)
                    + " (use preserve = ...)"
                    + " link ratio(s) is required for development estimation."
                )
            else:
                warning = (
                    "Some exclusions have been ignored. At least "
                    + str(preserve)
                    + " link ratio(s) is required for development estimation."
                )
            warnings.warn(warning)

        return weights.transpose((0, 1, 3, 2))

    def _drop_valuation(self, X):
        xp = X.get_array_module()
        if type(self.drop_valuation) is not list:
            drop_valuation = [self.drop_valuation]
        else:
            drop_valuation = self.drop_valuation
        v = pd.PeriodIndex(drop_valuation, freq=X.origin_grain).to_timestamp(how="e")
        arr = 1 - xp.nan_to_num(X[X.valuation.isin(v)].values[0, 0] * 0 + 1)
        ofill = X.shape[-2] - arr.shape[-2]
        dfill = X.shape[-1] - arr.shape[-1]
        if ofill > 0:
            arr = xp.concatenate(
                (arr, xp.repeat(xp.ones(arr.shape[-1])[None], ofill, 0)), 0
            )
        if dfill > 0:
            arr = xp.concatenate(
                (arr, xp.repeat(xp.ones(arr.shape[-2])[..., None], dfill, -1)), -1
            )
        return arr[:, :-1]

    def _drop(self, X):
        xp = X.get_array_module()
        drop = [self.drop] if type(self.drop) is not list else self.drop
        arr = X.nan_triangle.copy()
        for item in drop:
            arr[
                np.where(X.origin == item[0])[0][0],
                np.where(X.development == item[1])[0][0],
            ] = 0
        return arr[:, :-1]

    def _param_array_helper(self, size, param, default_value):
        # setting default
        param_array = pd.Series(size * [default_value]).astype("object")
        # only a single parameter is provided
        if isinstance(param, list):
            param_array[range(len(param))] = np.array(param)
            param_array.loc[param_array.isna()] = default_value
            param_array = param_array.astype(type(default_value))
        # an array of parameters is provided
        else:
            param_array.loc[:] = param
            param_array.loc[param_array.isna()] = default_value
            param_array = param_array.astype(type(default_value))
        return param_array.to_numpy()

    def _set_weight_func(self, factor, secondary_rank=None):
        w = (~np.isnan(factor.values)).astype(float)
        w = w * self._assign_n_periods_weight_func(factor)
        if self.drop is not None:
            w = w * self._drop_func(factor)

        if self.drop_valuation is not None:
            w = w * self._drop_valuation_func(factor)

        if (self.drop_above != np.inf) | (self.drop_below != 0.0):
            w = w * self._drop_x_func(factor)

        if (self.drop_high is not None) | (self.drop_low is not None):
            w = w * self._drop_n_func(factor * num_to_nan(w), secondary_rank)
        w_tri = factor.copy()
        w_tri.values = num_to_nan(w)
        return w_tri

    def _assign_n_periods_weight_func(self, factor):
        """Used to apply the n_periods weight"""
        # getting dimensions of factor for various manipulation
        factor_len = factor.shape[3]
        # putting n_periods into array
        n_periods_array = self._param_array_helper(factor_len, self.n_periods, -1)

        def _assign_n_periods_weight_int(X, n_periods):
            xp = X.get_array_module()
            val_offset = {
                "Y": {"Y": 1},
                "S": {"Y": 2, "S": 1},
                "Q": {"Y": 4, "Q": 1},
                "M": {"Y": 12, "Q": 3, "M": 1},
            }
            if n_periods < 1 or n_periods >= X.shape[-2]:
                return X.values * 0 + 1
            else:
                val_date_min = X.valuation[X.valuation <= X.valuation_date]
                val_date_min = val_date_min.drop_duplicates().sort_values()
                z = -n_periods * val_offset[X.development_grain][X.origin_grain]
                val_date_min = val_date_min[z]
                w = X[X.valuation >= val_date_min]
                return xp.nan_to_num((w / w).values) * X.nan_triangle

        xp = factor.get_array_module()

        dict_map = {
            item: _assign_n_periods_weight_int(factor, item)
            for item in set(n_periods_array)
        }
        conc = [
            dict_map[item][..., num : num + 1]
            for num, item in enumerate(n_periods_array)
        ]
        return xp.concatenate(tuple(conc), -1)

    def _drop_func(self, factor):
        # get the appropriate backend for nan_triangle and nan_to_num
        xp = factor.get_array_module()
        # turn single drop_valuation parameter to list if necessary
        drop_list = self.drop if isinstance(self.drop, list) else [self.drop]
        # get an starting array of weights
        arr = factor.nan_triangle.copy()
        # accommodate ldf triangle as factor, where the dimensions are '12-24'
        dev_list = (
            factor.development.str.split("-", expand=True)[0]
            if factor.development.dtype == object
            else factor.development.astype("string")
        )
        # create ndarray of drop_list for further operation in numpy
        drop_np = np.asarray(drop_list)
        # find indices of drop_np
        origin_ind = np.where(
            np.array([factor.origin.astype("string")]) == drop_np[:, [0]]
        )[1]
        dev_ind = np.where(np.array([dev_list]) == drop_np[:, [1]])[1]
        # set weight of dropped factors to 0
        arr[(origin_ind, dev_ind)] = 0
        return xp.nan_to_num(arr)[None, None]

    def _drop_valuation_func(self, factor):
        # get the appropriate backend for nan_to_num
        xp = factor.get_array_module()
        # turn single drop_valuation parameter to list if necessary
        if isinstance(self.drop_valuation, list):
            drop_valuation_list = self.drop_valuation
        else:
            drop_valuation_list = [self.drop_valuation]
        # turn drop_valuation to same valuation freq as factor
        v = pd.PeriodIndex(
            drop_valuation_list, freq=factor.development_grain
        ).to_timestamp(how="e")
        # warn that some drop_valuation are outside of factor
        if np.any(~v.isin(factor.valuation)):
            warnings.warn("Some valuations could not be dropped.")
        # return triangle of 0/1 where dropped factors have 0
        b = xp.nan_to_num(factor.iloc[0, 0][~factor.valuation.isin(v)].values * 0 + 1)
        # check to make sure some factors are still left
        if b.sum() == 0:
            raise Exception("The entire triangle has been dropped via drop_valuation.")
        return b

    def _drop_x_func(self, factor):
        # getting dimensions of factor for various manipulation
        factor_val = factor.values.copy()
        factor_len = factor_val.shape[3]
        indices = factor_val.shape[0]
        columns = factor_val.shape[1]

        # explicitly setting up 3D arrays for drop parameters to avoid broadcasting bugs
        drop_above_array = np.zeros((indices, columns, factor_len))
        drop_above_array[:, :, :] = self._param_array_helper(
            factor_len, self.drop_above, np.inf
        )[None, None]
        drop_below_array = np.zeros((indices, columns, factor_len))
        drop_below_array[:, :, :] = self._param_array_helper(
            factor_len, self.drop_below, 0.0
        )[None, None]
        preserve_array = np.zeros((indices, columns, factor_len))
        preserve_array[:, :, :] = self._param_array_helper(
            factor_len, self.preserve, self.preserve
        )[None, None]
        # transposing so columns of factors (same dev age) are in the last index.
        # not sure if this is really necessary. will leave for a better dev to find out
        factor_val_T = factor_val.transpose((0, 1, 3, 2))

        # setting up starting array of weights
        w = ~np.isnan(factor_val_T)

        # dropping
        index_array_weights = (factor_val_T < drop_above_array[..., None]) & (
            factor_val_T > drop_below_array[..., None]
        )

        # counting remaining factors
        ldf_count = index_array_weights.sum(axis=3)

        # applying preserve
        warning_flag = np.any(ldf_count < preserve_array)
        w = np.where(
            ldf_count[..., None] < preserve_array[..., None], w, index_array_weights
        )

        if warning_flag:
            if self.preserve == 1:
                warning = (
                    "Some exclusions have been ignored. At least "
                    + str(self.preserve)
                    + " (use preserve = ...)"
                    + " link ratio(s) is required for development estimation."
                )
            else:
                warning = (
                    "Some exclusions have been ignored. At least "
                    + str(self.preserve)
                    + " link ratio(s) is required for development estimation."
                )
            warnings.warn(warning)

        return w.transpose((0, 1, 3, 2)).astype(float)

    # for drop_high and drop_low
    def _drop_n_func(self, factor, secondary_rank=None):
        # getting dimensions of factor for various manipulation
        factor_val = factor.values.copy()
        # secondary rank is the optional triangle that breaks ties in factor
        # the original use case is for dropping the link ratio of 1 with the lowest loss value
        # (pass in a reverse rank of loss to drop link of ratio of 1 with the highest loss value)
        # leaving to user to ensure that secondary rank is the same dimensions as factor
        # also leaving to user to pick whether to trim head or tail
        if secondary_rank is None:
            sec_rank_val = factor_val.copy()
        else:
            sec_rank_val = secondary_rank.values.copy()
        factor_len = factor_val.shape[3]
        indices = factor_val.shape[0]
        columns = factor_val.shape[1]

        # explicitly setting up 3D arrays for drop parameters to avoid broadcasting bugs
        drop_high_array = np.zeros((indices, columns, factor_len))
        drop_high_array[:, :, :] = self._param_array_helper(
            factor_len, self.drop_high, 0
        )[None, None]
        drop_low_array = np.zeros((indices, columns, factor_len))
        drop_low_array[:, :, :] = self._param_array_helper(
            factor_len, self.drop_low, 0
        )[None, None]
        preserve_array = np.zeros((indices, columns, factor_len))
        preserve_array[:, :, :] = self._param_array_helper(
            factor_len, self.preserve, self.preserve
        )[None, None]

        # ranking factors by itself and secondary rank
        factor_ranks = np.lexsort((sec_rank_val, factor_val), axis=2).argsort(axis=2)

        # setting up starting weights
        w = ~np.isnan(factor_val.transpose((0, 1, 3, 2)))

        # counting valid factors
        ldf_count = w.sum(axis=3)

        # getting max index after drop high
        max_rank_unpreserve = ldf_count - drop_high_array

        # applying preserve
        preserve_trigger = (max_rank_unpreserve - drop_low_array) < preserve_array
        warning_flag = np.any(preserve_trigger)
        max_rank = np.where(preserve_trigger, ldf_count, max_rank_unpreserve)
        min_rank = np.where(preserve_trigger, 0, drop_low_array)

        # dropping
        index_array_weights = (
            factor_ranks.transpose((0, 1, 3, 2)) < max_rank[..., None]
        ) & (factor_ranks.transpose((0, 1, 3, 2)) > min_rank[..., None] - 1)

        if warning_flag:
            if self.preserve == 1:
                warning = (
                    "Some exclusions have been ignored. At least "
                    + str(self.preserve)
                    + " (use preserve = ...)"
                    + " link ratio(s) is required for development estimation."
                )
            else:
                warning = (
                    "Some exclusions have been ignored. At least "
                    + str(self.preserve)
                    + " link ratio(s) is required for development estimation."
                )
            warnings.warn(warning)

        return index_array_weights.transpose((0, 1, 3, 2)).astype(float)



import chainladder as cl
import pytest


def test_fit_period():
    tri = cl.load_sample("tail_sample")
    dev = cl.Development(average="simple").fit_transform(tri)
    assert (
        round(
            cl.TailCurve(fit_period=(tri.ddims[-7], None), extrap_periods=10)
            .fit(dev)
            .cdf_["paid"]
            .set_backend("numpy", inplace=True)
            .values[0, 0, 0, -2],
            3,
        )
        == 1.044
    )


def test_curve_validation():
    """
    Test validation of the curve parameter. Should raise a value error if an incorrect argument is supplied.
    """

    with pytest.raises(ValueError):
        tri = cl.load_sample('tail_sample')
        cl.TailCurve(
            curve='Exponential'
        ).fit_transform(tri)


def test_errors_validation():
    """
    Test validation of the errors parameter. Should raise a value error if an incorrect argument is supplied.
    """
    with pytest.raises(ValueError):
        tri = cl.load_sample('tail_sample')
        cl.TailCurve(
            errors='Ignore'
        ).fit_transform(tri)



import chainladder as cl
import numpy as np


def test_preserve_diagonal():
    triangle = cl.load_sample("berqsherm").loc["Auto"]
    xp = triangle.get_array_module()
    berq = cl.BerquistSherman(
        paid_amount="Paid",
        incurred_amount="Incurred",
        reported_count="Reported",
        closed_count="Closed",
    )
    berq_triangle = berq.fit_transform(triangle)
    assert (
        xp.nansum((berq_triangle.latest_diagonal - triangle.latest_diagonal).values)
        == 0
    )
    assert berq_triangle != triangle



import chainladder as cl
import numpy as np

def test_struhuss():
    X = cl.load_sample("cc_sample")["loss"]
    X = cl.TailConstant(tail=1 / 0.85).fit_transform(cl.Development().fit_transform(X))
    sample_weight = cl.load_sample("cc_sample")["exposure"].latest_diagonal
    ibnr = int(
        cl.CapeCod(trend=0.07, decay=0.75)
        .fit(X, sample_weight=sample_weight)
        .ibnr_.sum()
    )
    assert ibnr == 17052


def test_groupby(clrd):
    clrd = clrd[clrd['LOB']=='comauto']
    # But only the top 10 get their own CapeCod aprioris. Smaller companies get grouped together
    top_10 = clrd['EarnedPremDIR'].groupby('GRNAME').sum().latest_diagonal
    top_10 = top_10.loc[..., '1997', :].to_frame(origin_as_datetime=True).nlargest(10)
    cc_groupby = clrd.index['GRNAME'].map(lambda x: x if x in top_10.index else 'Remainder')
    idx = clrd.index
    idx['Top 10'] = cc_groupby
    clrd.index = idx

    # All companies share the same development factors regardless of size
    X = cl.Development().fit(clrd['CumPaidLoss'].sum()).transform(clrd['CumPaidLoss'])
    sample_weight=clrd['EarnedPremDIR'].latest_diagonal
    a = cl.CapeCod(groupby='Top 10', decay=0.98, trend=0.02).fit(X, sample_weight=sample_weight).ibnr_.groupby('Top 10').sum().sort_index()
    b = cl.CapeCod(decay=0.98, trend=0.02).fit(X.groupby('Top 10').sum(), sample_weight=sample_weight.groupby('Top 10').sum()).ibnr_.sort_index()
    xp = a.get_array_module()
    b = b.set_backend(a.array_backend)
    xp.allclose(xp.nan_to_num(a.values), xp.nan_to_num(b.values), atol=1e-5)


def test_capecod_zero_tri(raa):
    premium = raa.latest_diagonal * 0 + 50000
    raa.loc[:,:,'1987',48] = 0
    assert cl.CapeCod().fit(raa, sample_weight=premium).ultimate_.loc[:,:,'1987'].sum() > 0


def test_capecod_predict1(prism):
    """ github issue #400 
    Test whether we can make predictions at a more granular level than is fitted
    """
    prism = prism[['reportedCount', 'Paid']]

    cc_pipe = cl.Pipeline(
        [('dev', cl.Development()),
        ('model', cl.CapeCod())]
    )
    cc_pipe.fit(
        X=prism.groupby('Line')['Paid'].sum(), 
        sample_weight=prism.groupby('Line')['reportedCount'].sum().sum('development'))

    assert abs(cc_pipe.predict(prism['Paid'], sample_weight=prism['reportedCount'].sum('development')).ultimate_.sum() - 
            cc_pipe.named_steps.model.ultimate_.sum()).sum() < 1e-6
            

def test_capecod_predict2(prism):
    """ github issue #400 
    Test whether predictions between groupby with estimator and
    groupby outside estimator match
    """
    prism = prism[['reportedCount', 'Paid']]

    pipe1 = cl.Pipeline(
        [('dev', cl.Development(groupby='Line')),
        ('model', cl.CapeCod(groupby='Line'))]
    )
    pipe1.fit(
        X=prism['Paid'], 
        sample_weight=prism['reportedCount'].sum('development'))

    pipe2 = cl.Pipeline(
        [('dev', cl.Development()),
        ('model', cl.CapeCod())]
    )
    pipe2.fit(
        X=prism.groupby('Line')['Paid'].sum(), 
        sample_weight=prism.groupby('Line')['reportedCount'].sum().sum('development'))

    pred1 = pipe1.named_steps.model.ultimate_.sum()
    pred2 = pipe2.predict(prism['Paid'], sample_weight=prism['reportedCount'].sum('development')).ultimate_.sum()

    assert np.nan_to_num(abs(pred1 - pred2).values).sum() <= 1e-6


# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

from sklearn.base import BaseEstimator, TransformerMixin
from chainladder.core.io import EstimatorIO
import pandas as pd


class Trend(BaseEstimator, TransformerMixin, EstimatorIO):
    """
    Estimator to create and apply trend factors to a Triangle object.  Allows
    for compound trends as well as storage of the trend matrix to be used in
    other estimators, such as `CapeCod`.

    Parameters
    ----------

    trends: list-like
        The list containing the annual trends expressed as a decimal. For example,
        5% decrease should be stated as -0.05
    dates: list of date-likes
        A list-like of (start, end) dates to correspond to the `trend` list.
    axis: str (options: [‘origin’, ‘valuation’])
        The axis on which to apply the trend

    Attributes
    ----------

    trend_:
        A triangle representation of the trend factors

    """

    def __init__(self, trends=0.0, dates=None, axis="origin"):
        self.trends = trends
        self.dates = dates
        self.axis = axis

    def fit(self, X, y=None, sample_weight=None):
        """Fit the model with X.

        Parameters
        ----------
        X: Triangle-like
            Data to which the model will be applied.
        y: Ignored
        sample_weight: Ignored

        Returns
        -------
        self: object
            Returns the instance itself.
        """
        trends = self.trends if type(self.trends) is list else [self.trends]
        dates = [(None, None)] if self.dates is None else self.dates
        dates = [dates] if type(dates) is not list else dates
        if type(dates[0]) is not tuple:
            raise AttributeError(
                "Dates must be specified as a tuple of start and end dates"
            )
        self.trend_ = X.copy()
        for i, trend in enumerate(trends):
            self.trend_ = self.trend_.trend(
                trend, self.axis, start=dates[i][0], end=dates[i][1]
            )
        self.trend_ = self.trend_ / X
        return self

    def transform(self, X, y=None, sample_weight=None):
        """If X and self are of different shapes, align self to X, else
        return self.

        Parameters
        ----------
        X: Triangle
            The triangle to be transformed

        Returns
        -------
            X_new: New triangle with transformed attributes.
        """
        X_new = X.copy()
        triangles = ["trend_"]
        for item in triangles:
            setattr(X_new, item, getattr(self, item))
        X_new._set_slicers()
        return X_new


class TrendConstant(BaseEstimator, TransformerMixin, EstimatorIO):
    # """
    # Estimator to create and apply trend factors to a Triangle object.  Allows
    # for compound trends as well as storage of the trend matrix to be used in
    # other estimators, such as `CapeCod`.

    # Parameters
    # ----------

    # trends: list-like
    #     The list containing the annual trends expressed as a decimal. For example,
    #     5% decrease should be stated as -0.05
    # dates: list of date-likes
    #     A list-like of (start, end) dates to correspond to the `trend` list.
    # axis: str (options: [‘origin’, ‘valuation’])
    #     The axis on which to apply the trend

    # Attributes
    # ----------

    # trend_:
    #     A triangle representation of the trend factors

    # """

    def __init__(
        self,
        base_trend=0.0,
        trend_from="mid",
        trend_to_date=None,
        # dates=None,
        axis="origin",
    ):
        self.base_trend = base_trend
        self.trend_from = trend_from
        self.trend_to_date = trend_to_date
        self.axis = axis

    def fit(self, X, y=None, sample_weight=None):
        # """Fit the model with X.

        # Parameters
        # ----------
        # X: Triangle-like
        #     Data to which the model will be applied.
        # y: Ignored
        # sample_weight: Ignored

        # Returns
        # -------
        # self: object
        #     Returns the instance itself.
        # """
        print("IN TrendConstant FIT")
        print("base_trend", self.base_trend)

        self.trendedvalues_ = X.copy().trend(
            self.base_trend, self.axis  # , start=dates[i][0], end=dates[i][1]
        )
        print("self.trendedvalues_\n", self.trendedvalues_)

        # if type(dates[0]) is not tuple:
        #     raise AttributeError(
        #         "Dates must be specified as a tuple of start and end dates"
        #     )
        # self.trend_ = X.copy()
        # for i, trend in enumerate(trends):
        #     self.trend_ = self.trend_.trend(
        #         trend, self.axis, start=dates[i][0], end=dates[i][1]
        #     )
        self.trendfactor_ = self.trendedvalues_ / X
        print("self.trendfactor_\n", self.trendfactor_)
        return self

    def transform(self, X, y=None, sample_weight=None):
        # """ If X and self are of different shapes, align self to X, else
        # return self.

        # Parameters
        # ----------
        # X: Triangle
        #     The triangle to be transformed

        # Returns
        # -------
        #     X_new: New triangle with transformed attributes.
        # """
        X_new = X.copy()
        # triangles = ["trend_"]
        # for item in triangles:
        #     setattr(X_new, item, getattr(self, item))
        # X_new._set_slicers()
        return X_new



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
import numpy as np
import pandas as pd
import warnings
from sklearn.base import BaseEstimator
from chainladder.tails import TailConstant
from chainladder.development import Development
from chainladder.core.io import EstimatorIO
from chainladder.core.common import Common


class MethodBase(BaseEstimator, EstimatorIO, Common):

    _estimator_type = "chainladder"

    def validate_X(self, X):
        obj = X.copy()
        if "ldf_" not in obj:
            obj = Development().fit_transform(obj)
        if len(obj.ddims) - len(obj.ldf_.ddims) == 1:
            obj = TailConstant().fit_transform(obj)
        return obj.val_to_dev()

    def _align_cdf(self, X, sample_weight=None):
        """ Vertically align CDF to origin period latest diagonal. """
        valuation = X.valuation_date
        cdf = X.cdf_.iloc[..., : X.shape[-1]]
        a = X.iloc[0, 0] * 0
        a = a + a.nan_triangle
        if X.array_backend == "sparse":
            a = a - a[a.valuation < a.valuation_date]
        if sample_weight:
            X = X * a + sample_weight * a
        else:
            X = X * a
        cdf = X / X * cdf
        cdf.valuation_date = valuation
        return cdf.latest_diagonal

    def _set_ult_attr(self, ultimate):
        """ Ultimate scaffolding """
        from chainladder import options

        xp = ultimate.get_array_module()
        if ultimate.array_backend != "sparse":
            ultimate.values[~xp.isfinite(ultimate.values)] = xp.nan
        ultimate.ddims = pd.DatetimeIndex([options.ULT_VAL])
        ultimate.virtual_columns.columns = {}
        ultimate.is_cumulative = True
        ultimate._set_slicers()
        ultimate.valuation_date = ultimate.valuation.max()
        ultimate._drop_subtriangles()
        return ultimate

    @property
    def ldf_(self):
        return self.X_.ldf_

    @property
    def latest_diagonal(self):
        if self.X_.is_cumulative:
            return self.X_.latest_diagonal
        else:
            return self.X_.sum('development')

    def fit(self, X, y=None, sample_weight=None):
        """Applies the chainladder technique to triangle **X**

        Parameters
        ----------
        X : Triangle
            The data used to compute the mean and standard deviation
            used for later scaling along the features axis.
        y : Ignored
        sample_weight : Triangle
            For exposure-based methods, the exposure to be used for fitting

        Returns
        -------
        self : object
            Returns the instance itself.
        """
        self.X_ = self.validate_X(X)
        self.validate_weight(X, sample_weight)
        if sample_weight:
            self.sample_weight_ = sample_weight.set_backend(self.X_.array_backend)
        else:
            self.sample_weight_ = sample_weight
        return self

    def predict(self, X, sample_weight=None):
        """Predicts the chainladder ultimate on a new triangle **X**

        Parameters
        ----------
        X : Triangle
            The data used to compute the mean and standard deviation
            used for later scaling along the features axis.
        sample_weight : Triangle
            For exposure-based methods, the exposure to be used for predictions

        Returns
        -------
        X_new: Triangle

        """
        X_new = X.val_to_dev()
        if sum(X_new.ddims > self.ldf_.ddims.max()) > 0:
            raise ValueError("X has ages that exceed those available in model.")
        X_new = X_new + (self.X_.val_to_dev().iloc[0,0].sum(2) * 0)
        self.validate_weight(X_new, sample_weight)
        if sample_weight:
            sample_weight = sample_weight.set_backend(X_new.array_backend)
        X_new.ldf_ = self.ldf_
        X_new, X_new.ldf_ = self.intersection(X_new, X_new.ldf_)
        return X_new
        
    def intersection(self, a, b):
        """ Given two Triangles with mismatched indices, this method aligns
            their indices """
        if len(a) == 1 and len(b) == 1:
            return a, b
        intersection = list(set(a.key_labels).intersection(set(b.key_labels)))
        if intersection == []:
            return a, b
        a_idx = a.index[intersection]
        b_idx = b.index[intersection]
        idx_intersection = list(
            set(a_idx.set_index(intersection).index.intersection(
                b_idx.set_index(intersection).index)))
        if (len(a) == 1 or len(b) == 1) and idx_intersection == []:
            return a, b
        b = b.iloc[b_idx[b_idx[intersection].set_index(intersection).index.isin(idx_intersection)].index]
        a = a.iloc[a_idx[a_idx[intersection].set_index(intersection).index.isin(idx_intersection)].index]
        return a, b

    def fit_predict(self, X, y=None, sample_weight=None):
        self.fit(X, y, sample_weight)
        return self.predict(X, sample_weight)

    def _include_process_variance(self):
        if hasattr(self.X_, "_get_process_variance"):
            full = self.full_triangle_
            obj = self.X_._get_process_variance(full)
            self.ultimate_.values = obj.values[..., -1:]
            process_var = obj - full
        else:
            process_var = None
        return process_var

    def validate_weight(self, X, sample_weight):
        if (
            sample_weight
            and X.shape[:-1] != sample_weight.shape[:-1]
            and sample_weight.shape[2] != 1
            and sample_weight.shape[0] > 1
        ):
            warnings.warn(
                "X and sample_weight are not aligned. Broadcasting may occur.\n")



""" utils should store all utility functions and classes, i.e. things that
    are used by various modules in the package.
"""
from chainladder.utils.weighted_regression import (
    WeightedRegression,
)  # noqa (API import)

from chainladder.utils.utility_functions import (  # noqa (API import)
    parallelogram_olf,
    read_pickle,
    read_json,
    concat,
    load_sample,
    minimum,
    maximum,
    PatsyFormula,
    model_diagnostics
)
from chainladder.utils.cupy import cp
from chainladder.utils.sparse import sp
from chainladder.utils.dask import dp



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
import numpy as np
from chainladder import options

try:
    import dask.array as dp
    dp.array([1])
    module = "dask"
except:
    if options.ARRAY_BACKEND == "dask":
        import warnings

        warnings.warn("Unable to load Dask.  Using numpy instead.")
    import numpy as dp
    module = "numpy"

dp.nan = np.nan


def expand_dims(a, axis=0):
    l = []
    for i in range(len(a.shape)):
        if i == axis:
            l.append(None)
        l.append(slice(None))
    return a.__getitem__(tuple(l))

if dp != np:
    dp.expand_dims = expand_dims



from sklearn.utils import deprecated

class Options:
    ARRAY_BACKEND = "numpy"
    AUTO_SPARSE = True
    ARRAY_PRIORITY = ["dask", "sparse", "cupy", "numpy"]
    ULT_VAL = "2261-12-31 23:59:59.999999999"

    @classmethod
    def get_option(cls, option=None):
        return getattr(cls, option)

    @classmethod
    def set_option(cls, option, value):
        setattr(cls, option, value)

    def reset_option(self):
        self.set_option('ARRAY_BACKEND', "numpy")
        self.set_option('AUTO_SPARSE', True)
        self.set_option('ARRAY_PRIORITY', ["dask", "sparse", "cupy", "numpy"])
        self.set_option('ULT_VAL', "2261-12-31 23:59:59.999999999")

    def describe_option(self):
        pass

options = Options()

@deprecated("In an upcoming version of the package, this function will be deprecated. Use `chainladder.options.set_option('ARRAY_BACKEND', value)` to avoid breakage.")
def array_backend(array_backend="numpy"):
    options.set_option('ARRAY_BACKEND', array_backend)

@deprecated("In an upcoming version of the package, this function will be deprecated. Use `chainladder.options.set_option('AUTO_SPARSE', value)` to avoid breakage.")
def auto_sparse(auto_sparse=True):
    options.set_option('AUTO_SPARSE', auto_sparse)


from chainladder.utils import *  # noqa (API Import)
from chainladder.core import *  # noqa (API Import)
from chainladder.development import *  # noqa (API Import)
from chainladder.adjustments import *  # noqa (API Import)
from chainladder.tails import *  # noqa (API Import)
from chainladder.methods import *  # noqa (API Import)
from chainladder.workflow import *  # noqa (API Import)

__version__ = "0.8.23"



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import numpy as np
import pandas as pd
from chainladder import options
from chainladder.development.learning import DevelopmentML
from chainladder.development.glm import TweedieGLM
from sklearn.linear_model import LinearRegression
from sklearn.pipeline import Pipeline
from chainladder.utils.utility_functions import PatsyFormula


class BarnettZehnwirth(TweedieGLM):
    """ This estimator enables modeling from the Probabilistic Trend Family as
    described by Barnett and Zehnwirth.

    .. versionadded:: 0.8.2

    Parameters
    ----------
    formula: formula-like
        A patsy formula describing the independent variables, X of the GLM
    response:  str
        Column name for the reponse variable of the GLM.  If ommitted, then the
        first column of the Triangle will be used.


    """

    def __init__(self, formula='origin + development + valuation', response=None):
        self.formula = formula
        self.response = response

    def fit(self, X, y=None, sample_weight=None):
        if max(X.shape[:2]) > 1:
            raise ValueError("Only single index/column triangles are supported")
        tri = X.cum_to_incr().log()
        response = X.columns[0] if not self.response else self.response
        self.model_ = DevelopmentML(Pipeline(steps=[
            ('design_matrix', PatsyFormula(self.formula)),
            ('model', LinearRegression(fit_intercept=False))]),
                    y_ml=response, fit_incrementals=False).fit(tri)
        resid = tri - self.model_.triangle_ml_[
            self.model_.triangle_ml_.valuation <= tri.valuation_date]
        self.mse_resid_ = (resid**2).sum(0).sum(1).sum(2).sum() / (
            np.nansum(tri.nan_triangle) -
            len(self.model_.estimator_ml.named_steps.model.coef_))
        self.std_residuals_ = (resid / np.sqrt(self.mse_resid_))
        self.model_.triangle_ml_ = self.model_.triangle_ml_.exp()
        self.model_.triangle_ml_.is_cumulative = False
        return self

    def transform(self, X):
        """ If X and self are of different shapes, align self to X, else
        return self.

        Parameters
        ----------
        X : Triangle
            The triangle to be transformed

        Returns
        -------
            X_new : New triangle with transformed attributes.
        """
        X_new = X.copy()
        X_ml = self.model_._prep_X_ml(X.cum_to_incr().log())
        y_ml = self.model_.estimator_ml.predict(X_ml)
        triangle_ml = self.model_._get_triangle_ml(X_ml, y_ml)
        backend = "cupy" if X.array_backend == "cupy" else "numpy"
        triangle_ml.is_cumulative = False
        X_new.ldf_ = triangle_ml.exp().incr_to_cum().link_ratio.set_backend(backend)
        X_new.ldf_.valuation_date = pd.to_datetime(options.ULT_VAL)
        X_new._set_slicers()
        return X_new



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from chainladder.development.base import DevelopmentBase
import numpy as np
import pandas as pd
from scipy.optimize import minimize


class ClarkLDF(DevelopmentBase):
    """ A Estimator that allows for curve fitting development pattterns according
    to Clark 2003.

    The method fits incremental triangle amounts to one of
    ``loglogistic`` or ``weibull`` growth curves.  Both Clarks methods, LDF and Cape
    Cod, can be estimated.  To invoke the Cape Cod method, include
    ``sample_weight`` in when fitting the estimator.


    Parameters
    ----------
    growth: {'loglogistic', 'weibull'}
        The growth function to be used in curve fitting development patterns.
        Options are 'loglogistic' and 'weibull'
    groupby:
        An option to group levels of the triangle index together for the purposes
        of estimating patterns.  If omitted, each level of the triangle
        index will receive its own patterns.

    Attributes
    ----------
    ldf_: Triangle
        The estimated loss development patterns.
    cdf_: Triangle
        The estimated cumulative development patterns.
    incremental_fits_: Triangle
        The fitted incrementals of the model.
    theta_: DataFrame
        Estimates of the theta parameter of the growth curve.
    omega_: DataFrame
        Estimates of the omega parameter of the growth curve.
    elr_: DataFrame
        The Expected Loss Ratio parameter. This only exists when a ``sample_weight``
        is provided to the Estimator.
    scale_: DataFrame
        The scale parameter of the model.
    norm_resid_: Triangle
        The "Normalized" Residuals of the model according to Clark.

    """

    def __init__(self, growth="loglogistic", groupby=None):
        self.growth = growth
        self.groupby = groupby

    def _G(self, age, theta=None, omega=None):
        """ Growth function """
        xp = self.incremental_act_.get_array_module()
        if theta is None:
            theta = self.theta_.values[..., None, None]
        if omega is None:
            omega = self.omega_.values[..., None, None]
        age[age == 0.0] = xp.nan
        if self.growth == "loglogistic":
            out = 1 + (theta ** omega) * (age ** (-omega))
        elif self.growth == "weibull":
            out = 1 / (1 - xp.exp(-((age / theta) ** omega)))
        else:
            ValueError(str(self.growth) + "is an invalid growth curve.")
        out[xp.isnan(out)] = xp.inf
        return out

    def G_(self, age):
        """
        Growth function of the estimator.

        Parameters
        ----------
        age : int, float or array
            The age(s) at which to compute the value of the growth curve.

        Returns
        -------
        Triangle
            A Triangle object with growth curve values
        """
        xp = self.incremental_act_.get_array_module()
        if type(age) in [int, float, xp.int64, xp.float64]:
            age = xp.array([age]).astype("float64")
        if type(age) == list:
            age = xp.array([age]).astype("float64")
        obj = self.incremental_act_.copy()
        obj.odims = obj.odims[0:1]
        obj.values = 1 / self._G(age)
        obj.ddims = age
        return obj

    def fit(self, X, y=None, sample_weight=None):
        """Fit the model with X.

        Parameters
        ----------
        X : Triangle-like
            Set of LDFs to which the munich adjustment will be applied.
        y : Ignored
        sample_weight : Triangle-like
            Exposure vector used to invoke the Cape Cod method.

        Returns
        -------
        self : object
            Returns the instance itself.
        """
        from chainladder import options

        backend = X.array_backend
        if backend != "numpy":
            obj = X.set_backend("numpy", deep=True)
        else:
            obj = X.copy()
        obj = self._set_fit_groups(obj)
        X = self._set_fit_groups(X)
        xp = obj.get_array_module()
        nan_triangle = obj.nan_triangle
        ld = obj.latest_diagonal
        self.incremental_act_ = obj.cum_to_incr()
        if sample_weight:
            self.method_ = "cape_cod"
            sample_weight = self._set_fit_groups(sample_weight)
        else:
            self.method_ = "ldf"
        age_offset = {"Y": 6.0, "S": 3, "Q": 1.5, "M": 0.5}[X.development_grain]
        age_interval = {"Y": 12.0, "S": 6.0, "Q": 3.0, "M": 1.0}[X.development_grain]
        nans = nan_triangle.reshape(1, -1)[0]
        age = xp.tile(X.ddims, len(X.odims))[~xp.isnan(nans)].astype("float64")
        age_end = age - age_offset
        age_start = xp.maximum(age_end - age_interval, 0).astype("float64")
        origin = np.repeat(X.odims, len(X.ddims))[~xp.isnan(nans)]
        latest_diagonal = obj[X.valuation == X.valuation_date].sum("origin")
        latest_age = latest_diagonal.ddims.astype("float64")
        latest_origin = X.odims
        latest_diagonal = latest_diagonal.values[..., 0, :]

        index = xp.argsort(latest_origin)
        sorted_latest_origin = latest_origin[index]
        sorted_index = xp.searchsorted(sorted_latest_origin, origin)
        map_index = xp.take(index, sorted_index, mode="clip")

        increments = obj.cum_to_incr().values.reshape(X.shape[0], X.shape[1], -1, 1)[
            :, :, ~xp.isnan(nans), 0
        ]

        params = []
        for idx in range(len(X.index)):
            idx_params = []
            for col in range(len(X.columns)):

                def solver(x):
                    """ Solve Loglogistic MLE"""
                    ldf = lambda age: self._G(age, theta=x[..., 1], omega=x[..., 0])
                    if sample_weight:
                        ult = (
                            sample_weight.values[idx : idx + 1, col : col + 1, ::-1, 0]
                            * x[..., 2]
                        )
                    else:
                        ult = (
                            ldf(age=latest_age - age_offset)
                            * latest_diagonal[idx : idx + 1, col : col + 1]
                        )
                    increment_fit = ult[..., ::-1][..., map_index] * (
                        1 / ldf(age=age_end) - 1 / ldf(age=age_start)
                    )
                    mle = (
                        increments[idx : idx + 1, col : col + 1] * xp.log(increment_fit)
                        - increment_fit
                    )
                    return -xp.sum((xp.nan_to_num(mle.flatten())))

                if sample_weight:
                    x0 = xp.array([[[[1.0, age_interval, 1.0]]]])
                    bounds = ((1e-6, None), (1e-6, None), (1e-6, None))
                else:
                    x0 = xp.array([[[[1.0, age_interval]]]])
                    bounds = ((1e-6, None), (1e-6, None))
                idx_params.append(
                    minimize(fun=solver, x0=x0.flatten(), bounds=bounds).x.reshape(1, 1, 1, -1)
                )
            params.append(xp.concatenate(idx_params, axis=1))
        params = xp.concatenate(params, axis=0)
        cdf = self._G(
            latest_age - age_offset, theta=params[..., 1:2], omega=params[..., 0:1]
        )
        obj.values = cdf[..., :-1] / cdf[..., 1:]
        obj.ddims = X.link_ratio.ddims
        obj.odims = obj.odims[0:1]
        obj.is_pattern = True
        obj.is_cumulative = False
        obj._set_slicers()
        self.ldf_ = obj
        self.ldf_.valuation_date = pd.to_datetime(options.ULT_VAL)
        rows = X.index.set_index(X.key_labels).index
        self.omega_ = pd.DataFrame(params[..., 0, 0], index=rows, columns=X.vdims)
        self.theta_ = pd.DataFrame(params[..., 0, 1], index=rows, columns=X.vdims)
        if sample_weight:
            self.elr_ = pd.DataFrame(params[..., 0, 2], index=rows, columns=X.vdims)
        ultimate_ = (
            xp.swapaxes(self._G(age=(latest_age - age_offset)[::-1]), -1, -2)
            * ld.values
        )
        self.incremental_fits_ = X.copy()
        self.incremental_fits_.array_backend = "numpy"
        self.incremental_fits_.values = (
            (
                1 / self._G(X.ddims - age_offset)
                - 1 / self._G(xp.maximum(X.ddims - age_offset - age_interval, 0))
            )
            * ultimate_[..., ::-1]
            * nan_triangle
        )
        self.incremental_fits_.is_cumulative = False
        if backend == "cupy":
            self.set_backend("cupy", inplace=True)
        return self

    def transform(self, X):
        """ If X and self are of different shapes, align self to X, else
        return self.

        Parameters
        ----------
        X : Triangle
            The triangle to be transformed

        Returns
        -------
            X_new : New triangle with transformed attributes.
        """
        X_new = X.copy()
        triangles = [
            "ldf_",
            "omega_",
            "theta_",
            "incremental_fits_",
            "G_",
            "_G",
            "growth",
            "scale_",
        ]
        X_new.group_index = self._set_transform_groups(X_new)
        for item in triangles:
            setattr(X_new, item, getattr(self, item))
        X_new._set_slicers()
        if hasattr(self, "elr_"):
            X_new.elr_ = self.elr_
        return X_new

    @property
    def scale_(self):
        xp = self.incremental_fits_.get_array_module()
        scale = (
            (
                ((self.incremental_fits_ - self.incremental_act_) ** 2)
                / self.incremental_fits_
            )
            .sum("origin")
            .sum("development")
        )
        df = xp.nansum(self.incremental_fits_.nan_triangle) - 2
        if self.method_ == "ldf":
            df = df - len(self.incremental_fits_.odims)
        else:
            df = df - 1
        if scale.shape != ():
            scale = scale.values[..., 0, 0] / df
        else:
            scale = [[scale / df]]
        return pd.DataFrame(
            scale,
            index=self.incremental_fits_.index.set_index(
                self.incremental_fits_.key_labels
            ).index,
            columns=self.incremental_fits_.columns,
        )

    @property
    def norm_resid_(self):
        resid = (self.incremental_act_.values - self.incremental_fits_.values) / (
            np.sqrt(self.scale_.values[..., None, None] * self.incremental_fits_.values)
        )
        obj = self.incremental_fits_.copy()
        obj.values = resid
        return obj



### Building out a dev environment with a working copy
### of R ChainLadder is difficult.  These tests are 
### Currently inactive, but available should the compatibility
### of the installs improve at a later date.

import numpy as np
import pytest
import chainladder as cl

try:
    from rpy2.robjects.packages import importr
    from rpy2.robjects import r
    CL = importr("ChainLadder")
except:
    pass

def dev_corr_r(data, ci):
    return r("out<-dfCorTest({},ci={})".format(data, ci))


@pytest.mark.r
def val_corr_r(data, ci):
    return r("out<-cyEffTest({},ci={})".format(data, ci))


def val_corr_p(data, ci):
    return cl.load_sample(data).valuation_correlation(p_critical=ci, total=True)


data = ["RAA", "GenIns", "MW2014"]
ci = [0.5, 0.75]


@pytest.mark.r
@pytest.mark.parametrize("data", data)
@pytest.mark.parametrize("ci", ci)
def test_dev_corr(data, ci, atol):
    p = dev_corr_p(data, ci)
    r = dev_corr_r(data, ci).rx("T_stat")[0]
    p = p.t_expectation.values[0]
    assert np.allclose(r, p, atol=atol)


@pytest.mark.r
@pytest.mark.parametrize("data", data)
@pytest.mark.parametrize("ci", ci)
def test_dev_corr_var(data, ci, atol):
    p = dev_corr_p(data, ci)
    r = dev_corr_r(data, ci).rx("Var")[0]
    p = np.array([p.t_variance])
    assert np.allclose(r, p, atol=atol)


@pytest.mark.r
@pytest.mark.parametrize("data", data)
@pytest.mark.parametrize("ci", ci)
def test_dev_corr_range(data, ci, atol):
    p = dev_corr_p(data, 1 - ci)
    r = dev_corr_r(data, ci).rx("Range")[0]
    p = np.array(p.range)
    assert np.allclose(r, p, atol=atol)


@pytest.mark.r
@pytest.mark.parametrize("data", data)
@pytest.mark.parametrize("ci", ci)
def test_val_corr_z(data, ci, atol):
    p = val_corr_p(data, ci)
    r = val_corr_r(data, ci).rx("Z")[0]
    p = p.z.values[0]
    assert np.allclose(r, p, atol=atol)


@pytest.mark.r
@pytest.mark.parametrize("data", data)
@pytest.mark.parametrize("ci", ci)
def test_val_corr_e(data, ci, atol):
    p = val_corr_p(data, ci)
    r = val_corr_r(data, ci).rx("E")[0]
    p = p.z_expectation.values[0]
    assert np.allclose(r, p, atol=atol)


@pytest.mark.r
@pytest.mark.parametrize("data", data)
@pytest.mark.parametrize("ci", ci)
def test_val_corr_var(data, ci, atol):
    p = val_corr_p(data, ci)
    r = val_corr_r(data, ci).rx("Var")[0]
    p = p.z_variance.values[0]
    assert np.allclose(r, p, atol=atol)



import chainladder as cl

raa = cl.load_sample("RAA")
raa_1989 = raa[raa.valuation < raa.valuation_date]
cl_ult = cl.Chainladder().fit(raa).ultimate_  # Chainladder Ultimate
apriori = cl_ult * 0 + (float(cl_ult.sum()) / 10)  # Mean Chainladder Ultimate
apriori_1989 = apriori[apriori.origin < "1990"]


def test_cc_predict():
    cc = cl.CapeCod().fit(raa_1989, sample_weight=apriori_1989)
    assert cc.predict(raa, sample_weight=apriori)


def test_bf_predict():
    bf = cl.BornhuetterFerguson().fit(raa_1989, sample_weight=apriori_1989)
    assert bf.predict(raa, sample_weight=apriori)


def test_mack_predict():
    mack = cl.MackChainladder().fit(raa_1989)
    assert mack.predict(raa_1989)


def test_bs_random_state_predict(clrd):
    tri = clrd.groupby("LOB").sum().loc["wkcomp", ["CumPaidLoss", "EarnedPremNet"]]
    X = cl.BootstrapODPSample(random_state=100).fit_transform(tri["CumPaidLoss"])
    bf = cl.BornhuetterFerguson(apriori=0.6, apriori_sigma=0.1, random_state=42).fit(
        X, sample_weight=tri["EarnedPremNet"].latest_diagonal
    )
    assert (
        abs(
            bf.predict(X, sample_weight=tri["EarnedPremNet"].latest_diagonal)
            .ibnr_.sum()
            .sum()
            / bf.ibnr_.sum().sum()
            - 1
        )
        < 5e-3
    )


def test_basic_transform(raa):
    cl.Development().fit_transform(raa)
    cl.ClarkLDF().fit_transform(raa)
    cl.TailClark().fit_transform(raa)
    cl.TailBondy().fit_transform(raa)
    cl.TailConstant().fit_transform(raa)
    cl.TailCurve().fit_transform(raa)
    cl.BootstrapODPSample().fit_transform(raa)
    cl.IncrementalAdditive().fit_transform(raa, sample_weight=raa.latest_diagonal)


def test_misaligned_index(prism):
    prism = prism["Paid"]
    model = cl.Chainladder().fit(
        cl.Development(groupby=["Line", "Type"]).fit_transform(prism)
    )
    a = model.ultimate_.loc[prism.index.iloc[:10]].sum().sum()
    b = model.predict(prism.iloc[:10]).ultimate_.sum().sum()
    assert abs(a - b) < 1e-5


def test_misaligned_index2(clrd):
    clrd = clrd["CumPaidLoss"]
    w = cl.load_sample("clrd")["EarnedPremDIR"].latest_diagonal
    bcl = cl.Chainladder().fit(cl.Development(groupby=["LOB"]).fit_transform(clrd))
    bbk = cl.Benktander().fit(
        cl.Development(groupby=["LOB"]).fit_transform(clrd), sample_weight=w
    )
    bcc = cl.CapeCod().fit(
        cl.Development(groupby=["LOB"]).fit_transform(clrd), sample_weight=w
    )

    a = bcl.ultimate_.iloc[:10].sum().sum()
    b = bcl.predict(clrd.iloc[:10]).ultimate_.sum().sum()
    assert abs(a - b) < 1e-5
    a = bbk.ultimate_.iloc[:10].sum().sum()
    b = bbk.predict(clrd.iloc[:10], sample_weight=w.iloc[:10]).ultimate_.sum().sum()
    assert abs(a - b) < 1e-5
    a = bcc.ultimate_.iloc[:10].sum().sum()
    b = bcc.predict(clrd.iloc[:10], sample_weight=w.iloc[:10]).ultimate_.sum().sum()
    assert abs(a - b) < 1e-5

    a = bcl.ultimate_.iloc[150:153].sum().sum()
    b = bcl.predict(clrd.iloc[150:153]).ultimate_.sum().sum()
    assert abs(a - b) < 1e-5
    a = bbk.ultimate_.iloc[150:153].sum().sum()
    b = (
        bbk.predict(clrd.iloc[150:153], sample_weight=w.iloc[150:153])
        .ultimate_.sum()
        .sum()
    )
    assert abs(a - b) < 1e-5
    a = bcc.ultimate_.iloc[150:153].sum().sum()
    b = (
        bcc.predict(clrd.iloc[150:153], sample_weight=w.iloc[150:153])
        .ultimate_.sum()
        .sum()
    )
    assert abs(a - b) < 1e-5

    a = bcl.ultimate_.iloc[150:152].sum().sum()
    b = bcl.predict(clrd.iloc[150:152]).ultimate_.sum().sum()
    assert abs(a - b) < 1e-5
    a = bbk.ultimate_.iloc[150:152].sum().sum()
    b = (
        bbk.predict(clrd.iloc[150:152], sample_weight=w.iloc[150:152])
        .ultimate_.sum()
        .sum()
    )
    assert abs(a - b) < 1e-5
    a = bcc.ultimate_.iloc[150:152].sum().sum()
    b = (
        bcc.predict(clrd.iloc[150:152], sample_weight=w.iloc[150:152])
        .ultimate_.sum()
        .sum()
    )
    assert abs(a - b) < 1e-5

    a = bcl.ultimate_.iloc[150].sum().sum()
    b = bcl.predict(clrd.iloc[150]).ultimate_.sum().sum()
    assert abs(a - b) < 1e-5
    a = bbk.ultimate_.iloc[150].sum().sum()
    b = bbk.predict(clrd.iloc[150], sample_weight=w.iloc[150]).ultimate_.sum().sum()
    assert abs(a - b) < 1e-5
    a = bcc.ultimate_.iloc[150].sum().sum()
    b = bcc.predict(clrd.iloc[150], sample_weight=w.iloc[150]).ultimate_.sum().sum()
    assert abs(a - b) < 1e-5


def test_align_cdfs():
    ld = cl.load_sample("raa").latest_diagonal * 0 + 40000
    model = cl.BornhuetterFerguson().fit(cl.load_sample("raa"), sample_weight=ld)
    a = model.ultimate_.iloc[..., :4, :]
    b = model.predict(
        cl.load_sample("raa").dev_to_val().iloc[..., :4, -1].val_to_dev(),
        sample_weight=ld.iloc[..., :4, :],
    ).ultimate_
    assert a == b
    model = cl.Chainladder().fit(cl.load_sample("raa"), sample_weight=ld)
    a = model.ultimate_.iloc[..., :4, :]
    b = model.predict(
        cl.load_sample("raa").dev_to_val().iloc[..., :4, -1].val_to_dev(),
        sample_weight=ld.iloc[..., :4, :],
    ).ultimate_
    assert a == b


def test_check_val_tri_cl(raa):
    model = cl.Chainladder().fit(raa.dev_to_val())
    assert model.predict(raa.latest_diagonal).ultimate_ == model.ultimate_



import chainladder as cl


def test_bondy1():
    tri = cl.load_sample("tail_sample")["paid"]
    dev = cl.Development(average="simple").fit_transform(tri)
    assert round(float(cl.TailBondy(earliest_age=12).fit(dev).cdf_.values[0, 0, 0, -2]), 3) == 1.028



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from sklearn.model_selection import ParameterGrid
from sklearn.base import BaseEstimator
from sklearn.pipeline import Pipeline as PipelineSL
from sklearn.base import clone
from chainladder.core.io import EstimatorIO
from joblib import Parallel, delayed
import pandas as pd
import json


class GridSearch(BaseEstimator):
    """Exhaustive search over specified parameter values for an estimator.
    Important members are fit, predict.
    GridSearchCV implements a "fit" and a "score" method.
    It also implements "predict", "predict_proba", "decision_function",
    "transform" and "inverse_transform" if they are implemented in the
    estimator used.
    The parameters of the estimator used to apply these methods are optimized
    by cross-validated grid-search over a parameter grid.

    Parameters
    ----------
    estimator: estimator object.
        This is assumed to implement the chainladder estimator interface.
    param_grid: dict or list of dictionaries
        Dictionary with parameters names (string) as keys and lists of
        parameter settings to try as values, or a list of such
        dictionaries, in which case the grids spanned by each dictionary
        in the list are explored. This enables searching over any sequence
        of parameter settings.
    scoring: callable or dict of callable(s)
        Should be of the form {'name': callable}.  The callable(s) should
        return a single value.
    verbose: integer
        Controls the verbosity: the higher, the more messages.
    error_score: 'raise' or numeric
        Value to assign to the score if an error occurs in estimator fitting.
        If set to 'raise', the error is raised. If a numeric value is given,
        FitFailedWarning is raised. This parameter does not affect the refit
        step, which will always raise the error. Default is 'raise' but from
        version 0.22 it will change to np.nan.
    n_jobs: int, default=None
        The number of jobs to use for the computation. This will only provide
        speedup for n_targets > 1 and sufficient large problems.
        ``None`` means 1 unless in a :obj:`joblib.parallel_backend` context.
        ``-1`` means using all processors. See :term:`Glossary <n_jobs>`
        for more details.

    Attributes
    ----------
    results_: DataFrame
        A DataFrame with each param_grid key as a column and the ``scoring``
        score as the last column
    """

    def __init__(self, estimator, param_grid, scoring, verbose=0,
                 error_score="raise", n_jobs=None):
        self.estimator = estimator
        self.param_grid = param_grid
        self.scoring = scoring
        self.verbose = verbose
        self.error_score = error_score
        self.n_jobs = n_jobs

    def fit(self, X, y=None, **fit_params):
        """Fit the model with X.

        Parameters
        ----------
        X: Triangle-like
            Set of LDFs to which the tail will be applied.
        y: Ignored
        fit_params: (optional) dict of string -> object
            Parameters passed to the ``fit`` method of the estimator

        Returns
        -------
        self: object
            Returns the instance itself.
        """
        if type(self.scoring) is not dict:
            scoring = dict(score=self.scoring)
        else:
            scoring = self.scoring
        grid = list(ParameterGrid(self.param_grid))

        def _fit_single_estimator(estimator, fit_params, X, y, scoring, item):
            est = clone(estimator).set_params(**item)
            model = est.fit(X, y, **fit_params)
            for score in scoring.keys():
                item[score] = scoring[score](model)
            return item
            
        results_ = Parallel(n_jobs=self.n_jobs)(delayed(_fit_single_estimator)(
            self.estimator, fit_params, X, y, scoring, item)
            for item in grid)
        self.results_ = pd.DataFrame(results_)
        return self




class Pipeline(PipelineSL, EstimatorIO):
    """This is a near direct of copy the scikit-learn Pipeline class.

    Sequentially apply a list of transforms and a final estimator.
    Intermediate steps of the pipeline must be 'transforms', that is, they
    must implement fit and transform methods.
    The final estimator only needs to implement fit.
    The transformers in the pipeline can be cached using ``memory`` argument.
    The purpose of the pipeline is to assemble several steps that can be
    cross-validated together while setting different parameters.
    For this, it enables setting parameters of the various steps using their
    names and the parameter name separated by a '__', as in the example below.
    A step's estimator may be replaced entirely by setting the parameter
    with its name to another estimator, or a transformer removed by setting
    to None.
    Read more in the :ref:`User Guide <pipeline_docs>`.

    Parameters
    ----------
    steps: list
        List of (name, transform) tuples (implementing fit/transform) that are
        chained, in the order in which they are chained, with the last object
        an estimator.
    memory: None, str or object with the joblib.Memory interface, optional
        Used to cache the fitted transformers of the pipeline. By default,
        no caching is performed. If a string is given, it is the path to
        the caching directory. Enabling caching triggers a clone of
        the transformers before fitting. Therefore, the transformer
        instance given to the pipeline cannot be inspected
        directly. Use the attribute ``named_steps`` or ``steps`` to
        inspect estimators within the pipeline. Caching the
        transformers is advantageous when fitting is time consuming.
    Attributes
    ----------
    named_steps: bunch object, a dictionary with attribute access
        Read-only attribute to access any step parameter by user given name.
        Keys are step names and values are steps parameters."""

    def fit(self, X, y=None, sample_weight=None, **fit_params):
        if sample_weight:
            fit_params = {} if not fit_params else fit_params
            for step in self.steps:
                fit_params[step[0] + "__sample_weight"] = sample_weight
        return super().fit(X, y, **fit_params)

    def predict(self, X, sample_weight=None, **predict_params):
        if sample_weight:
            predict_params = {} if not predict_params else predict_params
            predict_params["sample_weight"] = sample_weight
        return super().predict(X, **predict_params)

    def fit_predict(self, X, y=None, sample_weight=None, **fit_params):
        self.fit(X, y, sample_weight, **fit_params)
        return self.predict(X, sample_weight, **fit_params)

    def to_json(self):
        return json.dumps(
            [
                {
                    "name": item[0],
                    "params": item[1].get_params(),
                    "__class__": item[1].__class__.__name__,
                }
                for item in self.steps
            ]
        )



import chainladder as cl
import pandas as pd


def test_parallelogram(clrd):
    lob = ["wkcomp"] * 3 + ["comauto"] * 3 + ["wkcomp"] * 2
    values = [0.05, 0.02, -0.1, 0.05, 0.05, 0.05, 0.2, 1 / 1.1 - 1]
    date = [
        "1/1/1989",
        "2/14/1990",
        "10/1/1992",
        "7/1/1988",
        "1/1/1990",
        "10/1/1993",
        "1/1/1996",
        "10/1/1992",
    ]
    rates = pd.DataFrame({"LOB": lob, "effdate": date, "change": values})

    olf = cl.ParallelogramOLF(
        rate_history=rates, change_col="change", date_col="effdate"
    )

    X = clrd["EarnedPremNet"].latest_diagonal
    X = X[X["LOB"].isin(["wkcomp", "comauto"])]
    X = olf.fit_transform(X)
    assert X.get_array_module().all(
        X.olf_.loc["comauto", "EarnedPremNet", "1994"].values
        - (9 / 12 * 9 / 12 / 2 * 0.05 + 1)
        < 0.005
    )
    assert X.get_array_module().all(
        X.olf_.loc["wkcomp", "EarnedPremNet", "1996"].values - 1.1 < 0.005
    )



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import numpy as np
import pandas as pd

from sklearn.preprocessing import OneHotEncoder, StandardScaler, PolynomialFeatures
from sklearn.compose import ColumnTransformer
from chainladder.development.base import DevelopmentBase
from chainladder import options


class DevelopmentML(DevelopmentBase):
    """ A Estimator that interfaces with machine learning (ML) tools that implement
    the scikit-learn API.

    The `DevelopmentML` estimator is used to generate ``ldf_`` patterns from
    the data.

    .. versionadded:: 0.8.1


    Parameters
    ----------
    estimator_ml: skearn Estimator
        Any sklearn compatible regression estimator, including Pipelines and
    y_ml: list or str or sklearn_transformer
        The response column(s) for the machine learning algorithm. It must be
        present within the Triangle.
    autoregressive: tuple, (autoregressive_col_name, lag, source_col_name)
        The subset of response column(s) to use as lagged features for the
        Time Series aspects of the model. Predictions from one development period
        get used as featues in the next development period. Lags should be negative
        integers.
    fit_incrementals:
        Whether the response variable should be converted to an incremental basis
        for fitting.

    Attributes
    ----------
    estimator_ml: Estimator
        An sklearn-style estimator to predict development patterns
    ldf_: Triangle
        The estimated loss development patterns.
    cdf_: Triangle
        The estimated cumulative development patterns.
    """

    def __init__(self, estimator_ml=None, y_ml=None, autoregressive=False,
                 weight_ml=None, fit_incrementals=True):
        self.estimator_ml=estimator_ml
        self.y_ml=y_ml
        self.weight_ml = weight_ml
        self.autoregressive=autoregressive
        self.fit_incrementals=fit_incrementals

    def _get_y_names(self):
        """ private function to get the response column name"""
        if not self.y_ml:
            y_names = self._columns
        if hasattr(self.y_ml, '_columns'):
            y_names = self.y_ml._columns
        elif isinstance(self.y_ml, ColumnTransformer):
            y_names = self.y_ml.transformers[0][-1]
        if type(self.y_ml) is list:
            y_names = self.y_ml
        elif type(self.y_ml) is str:
            y_names = [self.y_ml]
        return y_names


    @property
    def y_ml_(self):
        defaults = self._get_y_names()
        transformer = self.y_ml
        if not transformer:
            return ColumnTransformer(
                transformers=[('passthrough', 'passthrough', defaults)])
        elif type(transformer) is list:
            return ColumnTransformer(
                transformers=[('passthrough', 'passthrough', transformer)])
        elif type(transformer) is str:
            return ColumnTransformer(
                transformers=[('passthrough', 'passthrough', [transformer])])
        else:
            return transformer

    def _get_triangle_ml(self, df, preds=None):
        """ Create fitted Triangle """
        from chainladder.core import Triangle
        if preds is None:
            preds = self.estimator_ml.predict(df)
        X_r = [df]
        y_r = [preds]
        dgrain = {'Y':12, 'Q':3, 'M': 1, 'S': 6}[self.development_grain_]
        ograin = {'Y':1, 'Q':4, 'M': 12, 'S': 6}[self.origin_grain_]
        latest_filter = (df['origin']+1)*ograin+(df['development']-dgrain)/dgrain
        latest_filter = latest_filter == latest_filter.max()
        preds=pd.DataFrame(preds.copy())[latest_filter].values
        out = df.loc[latest_filter].copy()
        dev_lags = df['development'].drop_duplicates().sort_values()
        for d in dev_lags[1:]:
            out['development'] = out['development'] + dgrain
            out['valuation'] = out['valuation'] + dgrain / 12
            if len(preds.shape) == 1:
                preds = preds[:, None]
            if self.autoregressive:
                for num, col in enumerate(self.autoregressive):
                    out[col[0]]=preds[:, num]
            out = out[out['development']<=dev_lags.max()]
            if len(out) == 0:
                continue
            X_r.append(out.copy())
            preds = self.estimator_ml.predict(out)
            y_r.append(preds.copy())
        X_r = pd.concat(X_r, axis=0).reset_index(drop=True)
        if True:
            X_r = X_r.drop(self._get_y_names(), axis=1)
        out = pd.concat((X_r,
                         pd.DataFrame(np.concatenate(y_r, 0), columns=self._get_y_names())), axis=1)
        out['origin'] = out['origin'].map({v: k for k, v in self.origin_encoder_.items()})
        out['valuation'] = out['valuation'].map({v: k for k, v in self.valuation_encoder_.items()})
        return Triangle(
            out, origin='origin', development='valuation',
            index=self._key_labels, columns=self._get_y_names(),
            cumulative=not self.fit_incrementals).dropna()

    def _prep_X_ml(self, X):
        """ Preps Triangle data ahead of the pipeline """
        if self.fit_incrementals:
            X_ = X.cum_to_incr()
        else:
            X_ = X.copy()
        if self.autoregressive:
            for i in self.autoregressive:
                lag = X[i[2]].shift(i[1])
                X_[i[0]] = lag[lag.valuation<=X.valuation_date]
        df_base = X.incr_to_cum().to_frame(
            keepdims=True, implicit_axis=True, origin_as_datetime=True
            ).reset_index().iloc[:, :-1]
        df = df_base.merge(X.cum_to_incr().to_frame(
                keepdims=True, implicit_axis=True, origin_as_datetime=True
            ).reset_index(), how='left',
            on=list(df_base.columns)).fillna(0)
        df['origin'] = df['origin'].map(self.origin_encoder_)
        df['valuation'] = df['valuation'].map(self.valuation_encoder_)
        return df

    def fit(self, X, y=None, sample_weight=None):
        """Fit the model with X.

        Parameters
        ----------
        X : Triangle-like
            Set of LDFs to which the munich adjustment will be applied.
        y : None
            Ignored, use y_ml to set a reponse variable for the ML algorithm
        sample_weight : None
            Ignored

        Returns
        -------
        self : object
            Returns the instance itself.
        """

        self._columns = list(X.columns)
        self._key_labels = X.key_labels
        self.origin_grain_ = X.origin_grain
        self.development_grain_ = X.development_grain
        self.origin_encoder_ = dict(zip(
            X.origin.to_timestamp(how='s'),
            (pd.Series(X.origin).rank()-1)/{'Y':1, 'Q':4, 'M': 12, 'S': 6}[X.origin_grain]))
        val = X.valuation.sort_values().unique()
        self.valuation_encoder_ = dict(zip(
            val,
            (pd.Series(val).rank()-1)/{'Y':1, 'Q':4, 'M': 12, 'S': 6}[X.development_grain]))
        df = self._prep_X_ml(X)
        self.df_ = df
        # Fit model
        self.estimator_ml.fit(df, self.y_ml_.fit_transform(df).squeeze())
        #return self
        self.triangle_ml_ = self._get_triangle_ml(df)
        return self

    @property
    def ldf_(self):
        ldf = self.triangle_ml_.incr_to_cum().link_ratio
        ldf.valuation_date = pd.to_datetime(options.ULT_VAL)
        return ldf

    def transform(self, X):
        """ If X and self are of different shapes, align self to X, else
        return self.

        Parameters
        ----------
        X : Triangle
            The triangle to be transformed

        Returns
        -------
            X_new : New triangle with transformed attributes.
        """
        X_new = X.copy()
        X_ml = self._prep_X_ml(X)
        y_ml=self.estimator_ml.predict(X_ml)
        triangle_ml = self._get_triangle_ml(X_ml, y_ml)
        backend = "cupy" if X.array_backend == "cupy" else "numpy"
        X_new.ldf_ = triangle_ml.incr_to_cum().link_ratio.set_backend(backend)
        X_new.ldf_.valuation_date = pd.to_datetime(options.ULT_VAL)
        X_new._set_slicers()
        return X_new



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
import pandas as pd
from chainladder.utils.cupy import cp
from chainladder.utils.sparse import sp
from chainladder.utils.dask import dp
import numpy as np
from chainladder.utils.utility_functions import concat
from chainladder import options



def _get_full_expectation(cdf_, ultimate_, is_cumulative=True):
    """ Private method that builds full expectation"""
    full = ultimate_ / cdf_

    if is_cumulative:
        return concat((full, ultimate_.copy().rename("development", [9999])), axis=3)

    else:
        tail_ = full.iloc[:, :, :, -1] - ultimate_

        return concat(
            (full.cum_to_incr(), tail_.copy().rename("development", [9999])), axis=3
        )


def _get_full_triangle(X, ultimate, is_cumulative=True):
    """ Private method that builds full triangle"""
    # Getting the LDFs and expand for all origins
    from chainladder.utils.utility_functions import num_to_nan
    emergence = X.ldf_.copy() * (ultimate / ultimate)

    # Setting LDFs for all of the known diagonals as 1
    emergence = (
        emergence[emergence.valuation < X.valuation_date] * 0
        + 1
        + emergence[emergence.valuation >= X.valuation_date]
    )
    
    emergence.valuation_date = pd.to_datetime(options.ULT_VAL)
    emergence.values = emergence.values.cumprod(axis=3) - 1 
    
    # Shifting the CDFs by development age, and renaming the last column as 9999
    emergence.ddims = emergence.ddims + \
        {"Y": 12, "Q": 3, "S": 6, "M": 1}[emergence.development_grain]
    emergence.ddims[-1] = 9999
    emergence.values = emergence.values / num_to_nan(emergence.values[..., -1:])
    ld = X.incr_to_cum().latest_diagonal
    cum_run_off = ld + emergence * (ultimate - ld)
    cum_run_off = cum_run_off[cum_run_off.valuation > X.valuation_date]
    cum_run_off.is_cumulative = True

    if is_cumulative:
        return X + cum_run_off
    else:
        return (X.incr_to_cum() + cum_run_off).cum_to_incr()


class Common:
    """ Class that contains common properties of a "fitted" Triangle. """

    @property
    def has_ldf(self):
        if hasattr(self, "ldf_"):
            return True
        else:
            return False

    @property
    def has_zeta(self):
        if hasattr(self, "zeta_"):
            return True
        else:
            return False        
        
    @property
    def cdf_(self):
        if not self.has_ldf:
            x = self.__class__.__name__
            raise AttributeError("'" + x + "' object has no attribute 'cdf_'")
        return self.ldf_.incr_to_cum()

    @property
    def cum_zeta_(self):
        if not self.has_zeta:
            x = self.__class__.__name__
            raise AttributeError("'" + x + "' object has no attribute 'cum_zeta_'")
        return self.zeta_.incr_to_cum()
    
    @property
    def ibnr_(self):
        if not hasattr(self, "ultimate_"):
            x = self.__class__.__name__
            raise AttributeError("'" + x + "' object has no attribute 'ibnr_'")
        if hasattr(self, "X_"):
            ld = self.latest_diagonal
        else:
            ld = self.latest_diagonal if self.is_cumulative else self.sum(
                axis=3)
        ibnr = self.ultimate_ - ld
        ibnr.vdims = self.ultimate_.vdims
        return ibnr

    @property
    def full_expectation_(self):
        if not hasattr(self, "ultimate_"):
            raise AttributeError(
                "'"
                + self.__class__.__name__
                + "' object has no attribute 'full_expectation_'"
            )

        return _get_full_expectation(self.cdf_, self.ultimate_, self.X_.is_cumulative)

    @property
    def full_triangle_(self):
        if not hasattr(self, "ultimate_"):
            raise AttributeError(
                "'"
                + self.__class__.__name__
                + "' object has no attribute 'full_triangle_'"
            )

        if hasattr(self, "X_"):
            X = self.X_
        else:
            X = self
        return _get_full_triangle(X, self.ultimate_,  X.is_cumulative)
        

    def pipe(self, func, *args, **kwargs):
        return func(self, *args, **kwargs)

    def set_backend(self, backend, inplace=False, deep=False, **kwargs):
        """ Converts triangle array_backend.

        Parameters
        ----------
        backend : str
            Currently supported options are 'numpy', 'sparse', and 'cupy'
        inplace : bool
            Whether to mutate the existing Triangle instance or return a new
            one.

        Returns
        -------
            Triangle with updated array_backend
        """
        if hasattr(self, "array_backend"):
            old_backend = self.array_backend
        else:
            if hasattr(self, "ldf_"):
                old_backend = self.ldf_.array_backend
            else:
                raise ValueError("Unable to determine array backend.")
        if inplace:
            # Coming from dask - compute and then recall this method
            # going to dask  -
            if old_backend == "dask" and backend != "dask":
                self = self.compute()
                old_backend = self.array_backend
            if backend in ["numpy", "sparse", "cupy", "dask"]:
                lookup = {
                    "numpy": {
                        "sparse": lambda x: x.todense(),
                        "cupy": lambda x: cp.asnumpy(x),
                    },
                    "cupy": {
                        "numpy": lambda x: cp.array(x),
                        "sparse": lambda x: cp.array(x.todense()),
                    },
                    "sparse": {
                        "numpy": lambda x: sp.array(x),
                        "cupy": lambda x: sp.array(cp.asnumpy(x)),
                    },
                    "dask": {
                        # should this be chunked?
                        "numpy": lambda x: dp.from_array(x, **kwargs),
                        "cupy": lambda x: dp.from_array(x, **kwargs),
                        "sparse": lambda x: dp.from_array(x, **kwargs),
                    },
                }
                if hasattr(self, "values"):
                    self.values = lookup[backend].get(old_backend, lambda x: x)(
                        self.values
                    )
                if deep:
                    for k, v in vars(self).items():
                        if isinstance(v, Common):
                            v.set_backend(backend, inplace=True, deep=True)
                if hasattr(self, "array_backend"):
                    self.array_backend = backend
            else:
                raise AttributeError(backend, "backend is not supported.")
            return self
        else:
            obj = self.copy()
            return obj.set_backend(backend=backend, inplace=True, deep=deep, **kwargs)
    
    def _validate_assumption(self, triangle, value, axis):
        if type(value) in (int, float, str):
            arr = np.repeat(value, triangle.shape[axis])
        if type(value) in (list, tuple, set, np.ndarray):
            arr = np.array(value)
        if type(value) is dict:
            arr = np.array([value[a] for a in triangle._get_axis_value(axis)])
        if callable(value):
            arr = np.array([value(a) for a in triangle._get_axis_value(axis)])
        if axis == 3:
            arr = arr[None, None, None]
        if axis == 2:
            arr = arr[None, None, :, None]
        return arr



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
import pandas as pd
import numpy as np
from chainladder.development.base import DevelopmentBase
from chainladder.development.learning import DevelopmentML
from sklearn.linear_model import TweedieRegressor
from sklearn.pipeline import Pipeline
from chainladder.utils.utility_functions import PatsyFormula


class TweedieGLM(DevelopmentBase):
    """ This estimator creates development patterns with a GLM using a Tweedie distribution.

    The Tweedie family includes several of the more popular distributions including
    the normal, ODP poisson, and gamma distributions.  This class is a special case
    of `DevleopmentML`.  It restricts to just GLM using a TweedieRegressor and
    provides an R-like formulation of the design matrix.

    .. versionadded:: 0.8.1

    Parameters
    ----------
    design_matrix: formula-like
        A patsy formula describing the independent variables, X of the GLM
    response:  str
        Column name for the reponse variable of the GLM.  If ommitted, then the
        first column of the Triangle will be used.
    weight: str
        Column name of any weight to use in the GLM. If none specified, then an
        unweighted regression will be performed.
    power: float, default=0
            The power determines the underlying target distribution according
            to the following table:
            +-------+------------------------+
            | Power | Distribution           |
            +=======+========================+
            | 0     | Normal                 |
            +-------+------------------------+
            | 1     | Poisson                |
            +-------+------------------------+
            | (1,2) | Compound Poisson Gamma |
            +-------+------------------------+
            | 2     | Gamma                  |
            +-------+------------------------+
            | 3     | Inverse Gaussian       |
            +-------+------------------------+
            For ``0 < power < 1``, no distribution exists.
    alpha: float, default=1
        Constant that multiplies the penalty term and thus determines the
        regularization strength. ``alpha = 0`` is equivalent to unpenalized
        GLMs. In this case, the design matrix `X` must have full column rank
        (no collinearities).
    link: {'auto', 'identity', 'log'}, default='auto'
        The link function of the GLM, i.e. mapping from linear predictor
        `X @ coeff + intercept` to prediction `y_pred`. Option 'auto' sets
        the link depending on the chosen family as follows:
        - 'identity' for Normal distribution
        - 'log' for Poisson,  Gamma and Inverse Gaussian distributions
    max_iter: int, default=100
        The maximal number of iterations for the solver.
    tol: float, default=1e-4
        Stopping criterion. For the lbfgs solver,
        the iteration will stop when ``max{|g_j|, j = 1, ..., d} <= tol``
        where ``g_j`` is the j-th component of the gradient (derivative) of
        the objective function.
    warm_start: bool, default=False
        If set to ``True``, reuse the solution of the previous call to ``fit``
        as initialization for ``coef_`` and ``intercept_`` .
    verbose: int, default=0
        For the lbfgs solver set verbose to any positive number for verbosity.

    Attributes
    ----------
    model_: sklearn.Pipeline
        A scikit-learn Pipeline of the GLM
    """

    def __init__(self, design_matrix='C(development) + C(origin)',
                 response=None, weight=None, power=1.0, alpha=1.0, link='log',
                 max_iter=100, tol=0.0001, warm_start=False, verbose=0):
        self.response=response
        self.weight=weight
        self.design_matrix = design_matrix
        self.power=power
        self.alpha=alpha
        self.link=link
        self.max_iter=max_iter
        self.tol=tol
        self.warm_start=warm_start
        self.verbose=verbose

    def fit(self, X, y=None, sample_weight=None):
        response = X.columns[0] if not self.response else self.response
        self.model_ = DevelopmentML(Pipeline(steps=[
            ('design_matrix', PatsyFormula(self.design_matrix)),
            ('model', TweedieRegressor(
                    link=self.link, power=self.power, max_iter=self.max_iter,
                    tol=self.tol, warm_start=self.warm_start,
                    verbose=self.verbose, fit_intercept=False))]),
                    y_ml=response, weight_ml=self.weight).fit(X)
        return self

    @property
    def ldf_(self):
        return self.model_.ldf_

    @property
    def triangle_ml_(self):
        return self.model_.triangle_ml_

    @property
    def coef_(self):
        return pd.Series(
            self.model_.estimator_ml.named_steps.model.coef_, name='coef_',
            index=list(self.model_.estimator_ml.named_steps.design_matrix.
                            design_info_.column_name_indexes.keys())
        ).to_frame()

    def transform(self, X):
        return self.model_.transform(X)



import chainladder as cl
import pandas as pd
import numpy as np
import copy
import pytest


def test_grain(qtr):
    actual = qtr.iloc[0, 0].grain("OYDY")
    xp = actual.get_array_module()
    nan = xp.nan
    expected = xp.array(
        [
            [44, 621, 950, 1020, 1070, 1069, 1089, 1094, 1097, 1099, 1100, 1100],
            [42, 541, 1052, 1169, 1238, 1249, 1266, 1269, 1296, 1300, 1300, nan],
            [17, 530, 966, 1064, 1100, 1128, 1155, 1196, 1201, 1200, nan, nan],
            [10, 393, 935, 1062, 1126, 1209, 1243, 1286, 1298, nan, nan, nan],
            [13, 481, 1021, 1267, 1400, 1476, 1550, 1583, nan, nan, nan, nan],
            [2, 380, 788, 953, 1001, 1030, 1066, nan, nan, nan, nan, nan],
            [4, 777, 1063, 1307, 1362, 1411, nan, nan, nan, nan, nan, nan],
            [2, 472, 1617, 1818, 1820, nan, nan, nan, nan, nan, nan, nan],
            [3, 597, 1092, 1221, nan, nan, nan, nan, nan, nan, nan, nan],
            [4, 583, 1212, nan, nan, nan, nan, nan, nan, nan, nan, nan],
            [21, 422, nan, nan, nan, nan, nan, nan, nan, nan, nan, nan],
            [13, nan, nan, nan, nan, nan, nan, nan, nan, nan, nan, nan],
        ]
    )
    xp.testing.assert_array_equal(actual.values[0, 0, :, :], expected)


def test_grain_returns_valid_tri(qtr):
    assert qtr.grain("OYDY").latest_diagonal == qtr.latest_diagonal


def test_grain_increm_arg(qtr):
    clrd_i = qtr["incurred"].cum_to_incr()
    a = clrd_i.grain("OYDY").incr_to_cum()
    assert a == qtr["incurred"].grain("OYDY")


def test_commutative(qtr, atol):
    xp = qtr.get_array_module()
    full = cl.Chainladder().fit(qtr).full_expectation_
    assert qtr.grain("OYDY").val_to_dev() == qtr.val_to_dev().grain("OYDY")
    assert qtr.cum_to_incr().grain(
        "OYDY"
    ).val_to_dev() == qtr.val_to_dev().cum_to_incr().grain("OYDY")
    assert qtr.grain(
        "OYDY"
    ).cum_to_incr().val_to_dev().incr_to_cum() == qtr.val_to_dev().grain("OYDY")
    assert full.grain("OYDY").val_to_dev() == full.val_to_dev().grain("OYDY")
    assert full.cum_to_incr().grain(
        "OYDY"
    ).val_to_dev() == full.val_to_dev().cum_to_incr().grain("OYDY")
    a = full.grain("OYDY").cum_to_incr().val_to_dev().incr_to_cum()
    b = full.val_to_dev().grain("OYDY")
    assert abs(a - b).max().max().max() < atol


@pytest.mark.parametrize(
    "grain", ["OYDY", "OYDQ", "OYDM", "OSDS", "OSDQ", "OSDM", "OQDQ", "OQDM"]
)
@pytest.mark.parametrize("alt", [0, 1, 2])
@pytest.mark.parametrize("trailing", [False, True])
def test_different_forms_of_grain(prism_dense, grain, trailing, alt, atol):
    t = prism_dense["Paid"]
    if alt == 1:
        t = t.dev_to_val().copy()
    if alt == 2:
        t = t.val_to_dev().copy()
        t = t[t.valuation < "2017-09"]

    a = t.incr_to_cum().grain(grain, trailing=trailing)
    b = t.grain(grain, trailing=trailing).incr_to_cum()
    assert abs(a - b).sum().sum() < atol


def test_asymmetric_origin_grain(prism_dense):
    x = prism_dense.iloc[..., 8:, :].incr_to_cum()
    x = x[x.valuation < x.valuation_date]
    assert x.grain("OYDM").development[0] == 1

    x = x[x.valuation < x.valuation_date]
    assert x.grain("OYDM").development[0] == 1


def test_vector_triangle_grain_mismatch(prism):
    tri = prism["Paid"].sum().incr_to_cum().grain("OQDM")
    exposure = tri.latest_diagonal
    tri = tri.grain("OQDQ")
    assert (tri / exposure).development_grain == "Q"


def test_annual_trailing(prism):
    tri = prism["Paid"].sum().incr_to_cum()
    # (limit data to November)
    tri = tri[tri.valuation < tri.valuation_date].incr_to_cum()
    tri = tri.grain("OQDQ", trailing=True).grain("OYDY")
    assert np.all(tri.ddims[:4] == np.array([12, 24, 36, 48]))


def test_development_age():
    assert (
        cl.load_sample("raa").ddims == [12, 24, 36, 48, 60, 72, 84, 96, 108, 120]
    ).all()


def test_development_age_quarterly():
    assert (
        cl.load_sample("quarterly").ddims
        == [
            3,
            6,
            9,
            12,
            15,
            18,
            21,
            24,
            27,
            30,
            33,
            36,
            39,
            42,
            45,
            48,
            51,
            54,
            57,
            60,
            63,
            66,
            69,
            72,
            75,
            78,
            81,
            84,
            87,
            90,
            93,
            96,
            99,
            102,
            105,
            108,
            111,
            114,
            117,
            120,
            123,
            126,
            129,
            132,
            135,
        ]
    ).all()
